@@ -42,6 +42,7 @@ using namespace __gnu_cxx;
 
 #include "common/Mutex.h"
 #include "common/Timer.h"
+#include "common/Finisher.h"
 
 #include "common/compiler_extensions.h"
 
@@ -118,6 +119,7 @@ class MetaRequest;
 
 typedef void (*client_ino_callback_t)(void *handle, vinodeno_t ino, int64_t off, int64_t len);
 
+typedef int (*client_getgroups_callback_t)(void *handle, uid_t uid, gid_t **sgids);
 
 // ========================================================
 // client interface
@@ -175,6 +177,7 @@ struct dir_result_t {
 
   void reset() {
     last_name.clear();
+    at_cache_name.clear();
     next_offset = 2;
     this_offset = 0;
     offset = 0;
@@ -189,6 +192,14 @@ class Client : public Dispatcher {
 
   PerfCounters *logger;
 
+  class CommandHook : public AdminSocketHook {
+    Client *m_client;
+  public:
+    CommandHook(Client *client);
+    bool call(std::string command, std::string args, bufferlist& out);
+  };
+  CommandHook m_command_hook;
+
   // cluster descriptors
   MDSMap *mdsmap; 
   OSDMap *osdmap;
@@ -198,6 +209,11 @@ class Client : public Dispatcher {
   client_ino_callback_t ino_invalidate_cb;
   void *ino_invalidate_cb_handle;
 
+  client_getgroups_callback_t getgroups_cb;
+  void *getgroups_cb_handle;
+
+  Finisher async_ino_invalidator;
+
   Context *tick_event;
   utime_t last_cap_renew;
   void renew_caps();
@@ -227,6 +243,9 @@ public:
   map<tid_t, MetaRequest*> mds_requests;
   set<int>                 failed_mds;
 
+  void dump_mds_requests(Formatter *f);
+  void dump_mds_sessions(Formatter *f);
+
   int make_request(MetaRequest *req, int uid, int gid,
 		   //MClientRequest *req, int uid, int gid,
 		   Inode **ptarget = 0,
@@ -302,6 +321,16 @@ protected:
     free_fd_set.insert(fd, 1);
   }
 
+  /*
+   * Resolve file descriptor, or return NULL.
+   */
+  Fh *get_filehandle(int fd) {
+    hash_map<int, Fh*>::iterator p = fd_map.find(fd);
+    if (p == fd_map.end())
+      return NULL;
+    return p->second;
+  }
+
   // global client lock
   //  - protects Client and buffer cache both!
   Mutex                  client_lock;
@@ -318,6 +347,7 @@ protected:
   void close_dir(Dir *dir);
 
   friend class C_Client_PutInode; // calls put_inode()
+  friend class C_Client_CacheInvalidate;  // calls ino_invalidate_cb
 
   //int get_cache_size() { return lru.lru_get_size(); }
   //void set_cache_size(int m) { lru.lru_set_max(m); }
@@ -341,8 +371,8 @@ protected:
   void trim_dentry(Dentry *dn);
   void trim_caps(int mds, int max);
   
-  void dump_inode(Inode *in, set<Inode*>& did, bool disconnected);
-  void dump_cache();  // debug
+  void dump_inode(Formatter *f, Inode *in, set<Inode*>& did, bool disconnected);
+  void dump_cache(Formatter *f);  // debug
   
   // trace generation
   ofstream traceout;
@@ -396,7 +426,7 @@ protected:
   void flush_caps();
   void flush_caps(Inode *in, int mds);
   void kick_flushing_caps(int mds);
-  int get_caps(Inode *in, int need, int want, int *got, loff_t endoff);
+  int get_caps(Inode *in, int need, int want, int *have, loff_t endoff);
 
   void maybe_update_snaprealm(SnapRealm *realm, snapid_t snap_created, snapid_t snap_highwater, 
 			      vector<snapid_t>& snaps);
@@ -420,9 +450,11 @@ protected:
   void finish_cap_snap(Inode *in, CapSnap *capsnap, int used);
   void _flushed_cap_snap(Inode *in, snapid_t seq);
 
-  void _invalidate_inode_cache(Inode *in);
-  void _invalidate_inode_cache(Inode *in, int64_t off, int64_t len);
-  void _release(Inode *in, bool checkafter=true);
+  void _schedule_invalidate_callback(Inode *in, int64_t off, int64_t len, bool keep_caps);
+  void _invalidate_inode_cache(Inode *in, bool keep_caps);
+  void _invalidate_inode_cache(Inode *in, int64_t off, int64_t len, bool keep_caps);
+  void _async_invalidate(Inode *in, int64_t off, int64_t len, bool keep_caps);
+  void _release(Inode *in);
   bool _flush(Inode *in);
   void _flushed(Inode *in);
   void flush_set_callback(ObjectCacher::ObjectSet *oset);
@@ -436,6 +468,7 @@ protected:
   // metadata cache
   void update_dir_dist(Inode *in, DirStat *st);
 
+  void insert_readdir_results(MetaRequest *request, int mds, Inode *diri);
   Inode* insert_trace(MetaRequest *request, int mds);
   void update_inode_file_bits(Inode *in,
 			      uint64_t truncate_seq, uint64_t truncate_size, uint64_t size,
@@ -496,7 +529,9 @@ private:
   int _setxattr(Inode *in, const char *name, const void *value, size_t len, int flags, int uid=-1, int gid=-1);
   int _removexattr(Inode *in, const char *nm, int uid=-1, int gid=-1);
   int _open(Inode *in, int flags, mode_t mode, Fh **fhp, int uid=-1, int gid=-1);
-  int _create(Inode *in, const char *name, int flags, mode_t mode, Inode **inp, Fh **fhp, int uid=-1, int gid=-1);
+  int _create(Inode *in, const char *name, int flags, mode_t mode, Inode **inp, Fh **fhp,
+      int stripe_unit, int stripe_count, int object_size, const char *data_pool,
+	      int uid=-1, int gid=-1);
   loff_t _lseek(Fh *fh, loff_t offset, int whence);
   int _read(Fh *fh, int64_t offset, uint64_t size, bufferlist *bl);
   int _write(Fh *fh, int64_t offset, uint64_t size, const char *buf);
@@ -507,6 +542,8 @@ private:
   int get_or_create(Inode *dir, const char* name,
 		    Dentry **pdn, bool expect_null=false);
 
+  int check_permissions(Inode *in, int flags, int uid, int gid);
+
 public:
   int mount(const std::string &mount_root);
   void unmount();
@@ -569,18 +606,23 @@ public:
   int symlink(const char *existing, const char *newname);
 
   // inode stuff
+  int stat(const char *path, struct stat *stbuf, frag_info_t *dirstat=0, int mask=CEPH_STAT_CAP_INODE_ALL);
   int lstat(const char *path, struct stat *stbuf, frag_info_t *dirstat=0, int mask=CEPH_STAT_CAP_INODE_ALL);
   int lstatlite(const char *path, struct statlite *buf);
 
   int setattr(const char *relpath, struct stat *attr, int mask);
   int chmod(const char *path, mode_t mode);
+  int fchmod(int fd, mode_t mode);
   int chown(const char *path, uid_t uid, gid_t gid);
+  int fchown(int fd, uid_t uid, gid_t gid);
+  int lchown(const char *path, uid_t uid, gid_t gid);
   int utime(const char *path, struct utimbuf *buf);
   int truncate(const char *path, loff_t size);
 
   // file ops
   int mknod(const char *path, mode_t mode, dev_t rdev=0);
   int open(const char *path, int flags, mode_t mode=0);
+  int open(const char *path, int flags, mode_t mode, int stripe_unit, int stripe_count, int object_size, const char *data_pool);
   int lookup_hash(inodeno_t ino, inodeno_t dirino, const char *name);
   int lookup_ino(inodeno_t ino);
   int close(int fd);
@@ -658,6 +700,8 @@ public:
   int ll_statfs(vinodeno_t vino, struct statvfs *stbuf);
 
   void ll_register_ino_invalidate_cb(client_ino_callback_t cb, void *handle);
+
+  void ll_register_getgroups_cb(client_getgroups_callback_t cb, void *handle);
 };
 
 #endif
