@@ -146,6 +146,8 @@ public:
     raw *_raw;
     unsigned _off, _len;
 
+    void release();
+
   public:
     ptr() : _raw(0), _off(0), _len(0) {}
     ptr(raw *r);
@@ -162,7 +164,6 @@ public:
 
     raw *clone();
     void swap(ptr& other);
-    void release();
 
     // misc
     bool at_buffer_head() const { return _off == 0; }
@@ -197,6 +198,7 @@ public:
     unsigned wasted();
 
     int cmp(const ptr& o);
+    bool is_zero() const;
 
     // modifiers
     void set_offset(unsigned o) { _off = o; }
@@ -270,7 +272,7 @@ public:
 	//return off == bl->length();
       }
 
-      void advance(unsigned o);
+      void advance(int o);
       void seek(unsigned o);
       char operator*();
       iterator& operator++();
@@ -325,10 +327,13 @@ public:
 #endif
       return _len;
     }
+    bool contents_equal(buffer::list& other);
 
     bool is_page_aligned() const;
     bool is_n_page_sized() const;
 
+    bool is_zero() const;
+
     // modifiers
     void clear() {
       _buffers.clear();
@@ -366,6 +371,7 @@ public:
     // sort-of-like-assignment-op
     void claim(list& bl);
     void claim_append(list& bl);
+    void claim_prepend(list& bl);
 
     iterator begin() {
       return iterator(this, 0);
@@ -422,10 +428,30 @@ public:
     }
 
   };
+
+  /*
+   * efficient hash of one or more bufferlists
+   */
+
+  class hash {
+    __u32 crc;
+
+  public:
+    hash() : crc(0) { }
+
+    void update(buffer::list& bl) {
+      crc = bl.crc32c(crc);
+    }
+
+    __u32 digest() {
+      return crc;
+    }
+  };
 };
 
 typedef buffer::ptr bufferptr;
 typedef buffer::list bufferlist;
+typedef buffer::hash bufferhash;
 
 
 inline bool operator>(bufferlist& l, bufferlist& r) {
@@ -434,7 +460,6 @@ inline bool operator>(bufferlist& l, bufferlist& r) {
     if (l.length() == p) return false;
     if (l[p] > r[p]) return true;
     if (l[p] < r[p]) return false;
-    p++;
   }
 }
 inline bool operator>=(bufferlist& l, bufferlist& r) {
@@ -443,9 +468,18 @@ inline bool operator>=(bufferlist& l, bufferlist& r) {
     if (r.length() == p && l.length() == p) return true;
     if (l[p] > r[p]) return true;
     if (l[p] < r[p]) return false;
-    p++;
   }
 }
+
+inline bool operator==(bufferlist &l, bufferlist &r) {
+  if (l.length() != r.length())
+    return false;
+  for (unsigned p = 0; p < l.length(); p++) {
+    if (l[p] != r[p])
+      return false;
+  }
+  return true;
+}
 inline bool operator<(bufferlist& l, bufferlist& r) {
   return r > l;
 }
@@ -484,6 +518,11 @@ inline std::ostream& operator<<(std::ostream& out, buffer::error& e)
   return out << e.what();
 }
 
+inline bufferhash& operator<<(bufferhash& l, bufferlist &r) {
+  l.update(r);
+  return l;
+}
+
 }
 
 #endif
