@@ -71,7 +71,9 @@ librados::RadosClient::RadosClient(CephContext *cct_)
 {
 }
 
-int64_t librados::RadosClient::lookup_pool(const char *name) {
+int64_t librados::RadosClient::lookup_pool(const char *name)
+{
+  Mutex::Locker l(lock);
   int64_t ret = osdmap.lookup_pg_pool_name(name);
   if (ret < 0)
     return -ENOENT;
@@ -80,6 +82,7 @@ int64_t librados::RadosClient::lookup_pool(const char *name) {
 
 const char *librados::RadosClient::get_pool_name(int64_t pool_id)
 {
+  Mutex::Locker l(lock);
   return osdmap.get_pool_name(pool_id);
 }
 
@@ -178,12 +181,13 @@ int librados::RadosClient::connect()
   }
   messenger->set_myname(entity_name_t::CLIENT(monclient.get_global_id()));
 
+  objecter->init_unlocked();
   lock.Lock();
 
   timer.init();
 
   objecter->set_client_incarnation(0);
-  objecter->init();
+  objecter->init_locked();
   monclient.renew_subs();
 
   while (osdmap.get_epoch() == 0) {
@@ -217,11 +221,16 @@ void librados::RadosClient::shutdown()
     finisher.stop();
   }
   monclient.shutdown();
-  if (objecter && state == CONNECTED)
-    objecter->shutdown();
+  bool need_objecter = false;
+  if (objecter && state == CONNECTED) {
+    need_objecter = true;
+    objecter->shutdown_locked();
+  }
   state = DISCONNECTED;
   timer.shutdown();   // will drop+retake lock
   lock.Unlock();
+  if (need_objecter)
+    objecter->shutdown_unlocked();
   if (messenger) {
     messenger->shutdown();
     messenger->wait();
@@ -231,13 +240,10 @@ void librados::RadosClient::shutdown()
 
 uint64_t librados::RadosClient::get_instance_id()
 {
-  lock.Lock();
-  if (state == DISCONNECTED) {
-    lock.Unlock();
+  Mutex::Locker l(lock);
+  if (state == DISCONNECTED)
     return 0;
-  }
   uint64_t id = monclient.get_global_id();
-  lock.Unlock();
   return id;
 }
 
@@ -264,9 +270,9 @@ int librados::RadosClient::create_ioctx(const char *name, IoCtxImpl **io)
 
 bool librados::RadosClient::ms_dispatch(Message *m)
 {
+  Mutex::Locker l(lock);
   bool ret;
 
-  lock.Lock();
   if (state == DISCONNECTED) {
     ldout(cct, 10) << "disconnected, discarding " << *m << dendl;
     m->put();
@@ -274,7 +280,6 @@ bool librados::RadosClient::ms_dispatch(Message *m)
   } else {
     ret = _dispatch(m);
   }
-  lock.Unlock();
   return ret;
 }
 
