@@ -12,6 +12,8 @@
  * 
  */
 
+#include <unistd.h>
+
 #include "global/signal_handler.h"
 
 #include "include/types.h"
@@ -88,13 +90,13 @@ MDS::MDS(const std::string &n, Messenger *m, MonClient *mc) :
   mds_lock("MDS::mds_lock"),
   timer(m->cct, mds_lock),
   authorize_handler_cluster_registry(new AuthAuthorizeHandlerRegistry(m->cct,
-								      m->cct->_conf->auth_cluster_required.length() ?
-								      m->cct->_conf->auth_cluster_required :
-								      m->cct->_conf->auth_supported)),
+								      m->cct->_conf->auth_supported.length() ?
+								      m->cct->_conf->auth_supported :
+								      m->cct->_conf->auth_cluster_required)),
   authorize_handler_service_registry(new AuthAuthorizeHandlerRegistry(m->cct,
-								      m->cct->_conf->auth_service_required.length() ?
-								      m->cct->_conf->auth_service_required :
-								      m->cct->_conf->auth_supported)),
+								      m->cct->_conf->auth_supported.length() ?
+								      m->cct->_conf->auth_supported :
+								      m->cct->_conf->auth_service_required)),
   name(n),
   whoami(-1), incarnation(0),
   standby_for_rank(MDSMap::MDS_NO_STANDBY_PREF),
@@ -181,7 +183,7 @@ MDS::~MDS() {
     logger = 0;
   }
   if (mlogger) {
-    g_ceph_context->get_perfcounters_collection()->remove(logger);
+    g_ceph_context->get_perfcounters_collection()->remove(mlogger);
     delete mlogger;
     mlogger = 0;
   }
@@ -198,7 +200,7 @@ void MDS::create_logger()
 
     mds_plb.add_u64_counter(l_mds_req, "req"); // FIXME: nobody is actually setting this
     mds_plb.add_u64_counter(l_mds_reply, "reply");
-    mds_plb.add_fl_avg(l_mds_replyl, "replyl");
+    mds_plb.add_time_avg(l_mds_replyl, "replyl");
     mds_plb.add_u64_counter(l_mds_fw, "fw");
     
     mds_plb.add_u64_counter(l_mds_dir_f, "dir_f");
@@ -241,7 +243,7 @@ void MDS::create_logger()
     mds_plb.add_u64_counter(l_mds_trino, "trino");
     mds_plb.add_u64_counter(l_mds_tlock, "tlock");
     
-    mds_plb.add_fl(l_mds_l, "l");
+    mds_plb.add_u64(l_mds_l, "l");
     mds_plb.add_u64(l_mds_q, "q");
     mds_plb.add_u64(l_mds_popanyd, "popanyd"); // FIXME: unused
     mds_plb.add_u64(l_mds_popnest, "popnest");
@@ -411,7 +413,7 @@ void MDS::send_message_client_counted(Message *m, Session *session)
   if (session->connection) {
     messenger->send_message(m, session->connection);
   } else {
-    messenger->send_message(m, session->inst);
+    session->preopen_out_queue.push_back(m);
   }
 }
 
@@ -421,7 +423,7 @@ void MDS::send_message_client(Message *m, Session *session)
  if (session->connection) {
     messenger->send_message(m, session->connection);
   } else {
-    messenger->send_message(m, session->inst);
+    session->preopen_out_queue.push_back(m);
   }
 }
 
@@ -460,6 +462,8 @@ int MDS::init(int wanted_state)
   monc->authenticate();
   monc->wait_auth_rotating(30.0);
 
+  objecter->init_unlocked();
+
   mds_lock.Lock();
   if (want_state == CEPH_MDS_STATE_DNE) {
     mds_lock.Unlock();
@@ -502,8 +506,8 @@ int MDS::init(int wanted_state)
   beacon_start();
   whoami = -1;
   messenger->set_myname(entity_name_t::MDS(whoami));
-
-  objecter->init();
+  
+  objecter->init_locked();
 
   monc->sub_want("mdsmap", 0, 0);
   monc->renew_subs();
@@ -557,7 +561,7 @@ void MDS::tick()
   if (logger) {
     req_rate = logger->get(l_mds_req);
     
-    logger->fset(l_mds_l, (int)load.mds_load());
+    logger->set(l_mds_l, 100 * load.mds_load());
     logger->set(l_mds_q, messenger->get_dispatch_queue_len());
     logger->set(l_mds_sm, mdcache->num_subtrees());
 
@@ -1229,10 +1233,10 @@ void MDS::starting_done()
   assert(is_starting());
   request_state(MDSMap::STATE_ACTIVE);
 
+  mdcache->open_root();
+
   // start new segment
   mdlog->start_new_segment(0);
-
-  mdcache->open_root();
 }
 
 
@@ -1413,6 +1417,7 @@ void MDS::reconnect_done()
 void MDS::rejoin_joint_start()
 {
   dout(1) << "rejoin_joint_start" << dendl;
+  mdcache->finish_committed_masters();
   mdcache->rejoin_send_rejoins();
 }
 void MDS::rejoin_done()
@@ -1557,7 +1562,7 @@ void MDS::suicide()
   // shut down cache
   mdcache->shutdown();
 
-  objecter->shutdown();
+  objecter->shutdown_locked();
   
   // shut down messenger
   messenger->shutdown();
@@ -1579,17 +1584,10 @@ void MDS::respawn()
   }
   new_argv[orig_argc] = NULL;
 
-#if defined(__linux__)
-  dout(1) << " cwd " << get_current_dir_name() << dendl;
-#else
-  {
-#include <sys/param.h>
-    char buf[PATH_MAX];
-    
-    getcwd(buf, sizeof(buf));
-    dout(1) << " cwd " << buf << dendl;
-  }
-#endif
+  char buf[PATH_MAX];
+  char *cwd = getcwd(buf, sizeof(buf));
+  assert(cwd);
+  dout(1) << " cwd " << cwd << dendl;
 
   unblock_all_signals(NULL);
   execv(orig_argv[0], new_argv);
@@ -1713,9 +1711,10 @@ bool MDS::handle_deferrable_message(Message *m)
     switch (m->get_type()) {
       // SERVER
     case CEPH_MSG_CLIENT_SESSION:
-    case CEPH_MSG_CLIENT_REQUEST:
     case CEPH_MSG_CLIENT_RECONNECT:
       ALLOW_MESSAGES_FROM(CEPH_ENTITY_TYPE_CLIENT);
+      // fall-thru
+    case CEPH_MSG_CLIENT_REQUEST:
       server->dispatch(m);
       break;
     case MSG_MDS_SLAVE_REQUEST:
@@ -2016,7 +2015,7 @@ void MDS::ms_handle_remote_reset(Connection *con)
 
 bool MDS::ms_verify_authorizer(Connection *con, int peer_type,
 			       int protocol, bufferlist& authorizer_data, bufferlist& authorizer_reply,
-			       bool& is_valid)
+			       bool& is_valid, CryptoKey& session_key)
 {
   Mutex::Locker l(mds_lock);
   if (want_state == CEPH_MDS_STATE_DNE)
@@ -2041,7 +2040,7 @@ bool MDS::ms_verify_authorizer(Connection *con, int peer_type,
   uint64_t global_id;
 
   is_valid = authorize_handler->verify_authorizer(cct, monc->rotating_secrets,
-						  authorizer_data, authorizer_reply, name, global_id, caps_info);
+						  authorizer_data, authorizer_reply, name, global_id, caps_info, session_key);
 
   if (is_valid) {
     // wire up a Session* to this connection, and add it to the session map
@@ -2097,6 +2096,12 @@ void MDS::ms_handle_accept(Connection *con)
     if (s->connection != con) {
       dout(10) << " session connection " << s->connection << " -> " << con << dendl;
       s->connection = con;
+
+      // send out any queued messages
+      while (!s->preopen_out_queue.empty()) {
+	messenger->send_message(s->preopen_out_queue.front(), con);
+	s->preopen_out_queue.pop_front();
+      }
     }
   }
 }
