@@ -16,6 +16,7 @@
 #include <sstream>
 #include <stdlib.h>
 #include <signal.h>
+#include <limits.h>
 
 #include "Monitor.h"
 #include "common/version.h"
@@ -80,16 +81,29 @@ static ostream& _prefix(std::ostream *_dout, const Monitor *mon) {
 		<< "(" << mon->get_state_name() << ") e" << mon->monmap->get_epoch() << " ";
 }
 
-CompatSet get_ceph_mon_feature_compat_set()
+long parse_pos_long(const char *s, ostream *pss)
 {
-  CompatSet::FeatureSet ceph_mon_feature_compat;
-  CompatSet::FeatureSet ceph_mon_feature_ro_compat;
-  CompatSet::FeatureSet ceph_mon_feature_incompat;
-  ceph_mon_feature_incompat.insert(CEPH_MON_FEATURE_INCOMPAT_BASE);
-  return CompatSet(ceph_mon_feature_compat, ceph_mon_feature_ro_compat,
-		   ceph_mon_feature_incompat);
+  if (*s == '-' || *s == '+') {
+    *pss << "expected numerical value, got: " << s;
+    return -EINVAL;
+  }
+
+  string err;
+  long r = strict_strtol(s, 10, &err);
+  if ((r == 0) && !err.empty()) {
+    if (pss)
+      *pss << err;
+    return -1;
+  }
+  if (r < 0) {
+    if (pss)
+      *pss << "unable to parse positive integer '" << s << "'";
+    return -1;
+  }
+  return r;
 }
 
+
 Monitor::Monitor(CephContext* cct_, string nm, MonitorStore *s, Messenger *m, MonMap *map) :
   Dispatcher(cct_),
   name(nm),
@@ -103,11 +117,11 @@ Monitor::Monitor(CephContext* cct_, string nm, MonitorStore *s, Messenger *m, Mo
   clog(cct_, messenger, monmap, LogClient::FLAG_MON),
   key_server(cct, &keyring),
   auth_cluster_required(cct,
-			cct->_conf->auth_cluster_required.length() ?
-			cct->_conf->auth_cluster_required : cct->_conf->auth_supported),
+			cct->_conf->auth_supported.length() ?
+			cct->_conf->auth_supported : cct->_conf->auth_cluster_required),
   auth_service_required(cct,
-		       cct->_conf->auth_service_required.length() ?
-		       cct->_conf->auth_service_required : cct->_conf->auth_supported),
+			cct->_conf->auth_supported.length() ?
+			cct->_conf->auth_supported : cct->_conf->auth_service_required),
   store(s),
   
   state(STATE_PROBING),
@@ -180,22 +194,114 @@ Monitor::~Monitor()
     delete *p;
   for (vector<Paxos*>::iterator p = paxos.begin(); p != paxos.end(); p++)
     delete *p;
-  //clean out MonSessionMap's subscriptions
-  for (map<string, xlist<Subscription*>* >::iterator i
-	 = session_map.subs.begin();
-       i != session_map.subs.end();
-       ++i) {
-    while (!i->second->empty())
-      session_map.remove_sub(i->second->front());
-    delete i->second;
-  }
-  //clean out MonSessionMap's sessions
-  while (!session_map.sessions.empty()) {
-    session_map.remove_session(session_map.sessions.front());
-  }
+  assert(session_map.sessions.empty());
   delete mon_caps;
 }
 
+void Monitor::recovered_leader(int id)
+{
+  dout(10) << "recovered_leader " << id << " " << get_paxos_name(id) << " (" << paxos_recovered << ")" << dendl;
+  assert(paxos_recovered.count(id) == 0);
+  paxos_recovered.insert(id);
+  if (paxos_recovered.size() == paxos.size()) {
+    dout(10) << "all paxos instances recovered, going writeable" << dendl;
+
+    if (!features.incompat.contains(CEPH_MON_FEATURE_INCOMPAT_GV) &&
+	(quorum_features & CEPH_FEATURE_MON_GV)) {
+      require_gv_ondisk();
+      require_gv_onwire();
+    }
+
+    for (vector<Paxos*>::iterator p = paxos.begin(); p != paxos.end(); p++) {
+      if (!(*p)->is_active())
+	continue;
+      finish_contexts(g_ceph_context, (*p)->waiting_for_active);
+    }
+    for (vector<Paxos*>::iterator p = paxos.begin(); p != paxos.end(); p++) {
+      if (!(*p)->is_active())
+	continue;
+      finish_contexts(g_ceph_context, (*p)->waiting_for_commit);
+    }
+    for (vector<Paxos*>::iterator p = paxos.begin(); p != paxos.end(); p++) {
+      if (!(*p)->is_readable())
+	continue;
+      finish_contexts(g_ceph_context, (*p)->waiting_for_readable);
+    }
+    for (vector<Paxos*>::iterator p = paxos.begin(); p != paxos.end(); p++) {
+      if (!(*p)->is_writeable())
+	continue;
+      finish_contexts(g_ceph_context, (*p)->waiting_for_writeable);
+    }
+  }
+}
+
+void Monitor::recovered_peon(int id)
+{
+  // unlike recovered_leader(), recovered_peon() can get called
+  // multiple times, because it is triggered by a paxos lease message,
+  // and the leader may send multiples of those out for a given paxos
+  // machine while it is waiting for another instance to recover.
+  if (paxos_recovered.count(id))
+    return;
+  dout(10) << "recovered_peon " << id << " " << get_paxos_name(id) << " (" << paxos_recovered << ")" << dendl;
+  paxos_recovered.insert(id);
+  if (paxos_recovered.size() == paxos.size()) {
+    dout(10) << "all paxos instances recovered/leased" << dendl;
+
+    if (!features.incompat.contains(CEPH_MON_FEATURE_INCOMPAT_GV) &&
+	(quorum_features & CEPH_FEATURE_MON_GV)) {
+      require_gv_ondisk();
+      require_gv_onwire();
+    }
+  }
+}
+
+void Monitor::require_gv_ondisk()
+{
+  dout(0) << "setting CEPH_MON_FEATURE_INCOMPAT_GV" << dendl;
+  features.incompat.insert(CEPH_MON_FEATURE_INCOMPAT_GV);
+  write_features();
+}
+
+void Monitor::require_gv_onwire()
+{
+  dout(10) << "require_gv_onwire" << dendl;
+  // require protocol feature bit of my peers
+  Messenger::Policy p = messenger->get_policy(entity_name_t::TYPE_MON);
+  p.features_required |= CEPH_FEATURE_MON_GV;
+  messenger->set_policy(entity_name_t::TYPE_MON, p);
+}
+
+version_t Monitor::get_global_paxos_version()
+{
+  // this should only be called when paxos becomes writeable, which is
+  // *after* everything settles after an election.
+  assert(is_all_paxos_recovered());
+
+  if ((quorum_features & CEPH_FEATURE_MON_GV) == 0) {
+    // do not sure issuing gv's until the entire quorum supports them.
+    // this way we synchronize the setting of the incompat GV ondisk
+    // feature with actually writing the values to the data store, and
+    // avoid having to worry about hybrid cases.
+    dout(10) << "get_global_paxos_version no-op; quorum does not support the feature" << dendl;
+    return 0;
+  }
+
+  if (global_version == 0) {
+    global_version =
+      osdmon()->paxos->get_version() +
+      mdsmon()->paxos->get_version() +
+      monmon()->paxos->get_version() +
+      pgmon()->paxos->get_version() +
+      authmon()->paxos->get_version() +
+      logmon()->paxos->get_version();
+    dout(10) << "get_global_paxos_version first call this election epoch, starting from " << global_version << dendl;
+  }
+  ++global_version;
+  dout(20) << "get_global_paxos_version " << global_version << dendl;
+  return global_version;
+}
+
 enum {
   l_mon_first = 456000,
   l_mon_last,
@@ -234,11 +340,85 @@ void Monitor::handle_signal(int signum)
   shutdown();
 }
 
-int Monitor::init()
+CompatSet Monitor::get_supported_features()
+{
+  CompatSet::FeatureSet ceph_mon_feature_compat;
+  CompatSet::FeatureSet ceph_mon_feature_ro_compat;
+  CompatSet::FeatureSet ceph_mon_feature_incompat;
+  ceph_mon_feature_incompat.insert(CEPH_MON_FEATURE_INCOMPAT_BASE);
+  ceph_mon_feature_incompat.insert(CEPH_MON_FEATURE_INCOMPAT_GV);
+  return CompatSet(ceph_mon_feature_compat, ceph_mon_feature_ro_compat,
+		   ceph_mon_feature_incompat);
+}
+
+CompatSet Monitor::get_legacy_features()
+{
+  CompatSet::FeatureSet ceph_mon_feature_compat;
+  CompatSet::FeatureSet ceph_mon_feature_ro_compat;
+  CompatSet::FeatureSet ceph_mon_feature_incompat;
+  ceph_mon_feature_incompat.insert(CEPH_MON_FEATURE_INCOMPAT_BASE);
+  return CompatSet(ceph_mon_feature_compat, ceph_mon_feature_ro_compat,
+		   ceph_mon_feature_incompat);
+}
+
+int Monitor::check_features(MonitorStore *store)
+{
+  CompatSet required = get_supported_features();
+  CompatSet ondisk;
+
+  bufferlist features;
+  store->get_bl_ss_safe(features, COMPAT_SET_LOC, 0);
+  if (features.length() == 0) {
+    generic_dout(0) << "WARNING: mon fs missing feature list.\n"
+	    << "Assuming it is old-style and introducing one." << dendl;
+    //we only want the baseline ~v.18 features assumed to be on disk.
+    //If new features are introduced this code needs to disappear or
+    //be made smarter.
+    ondisk = get_legacy_features();
+
+    bufferlist bl;
+    ondisk.encode(bl);
+    store->put_bl_ss(bl, COMPAT_SET_LOC, 0);
+  } else {
+    bufferlist::iterator it = features.begin();
+    ondisk.decode(it);
+  }
+
+  if (!required.writeable(ondisk)) {
+    CompatSet diff = required.unsupported(ondisk);
+    generic_derr << "ERROR: on disk data includes unsupported features: " << diff << dendl;
+    return -EPERM;
+  }
+
+  return 0;
+}
+
+void Monitor::read_features()
+{
+  bufferlist bl;
+  store->get_bl_ss_safe(bl, COMPAT_SET_LOC, 0);
+  assert(bl.length());
+
+  bufferlist::iterator p = bl.begin();
+  ::decode(features, p);
+  dout(10) << "features " << features << dendl;
+
+  if (features.incompat.contains(CEPH_MON_FEATURE_INCOMPAT_GV))
+    require_gv_onwire();
+}
+
+void Monitor::write_features()
+{
+  bufferlist bl;
+  features.encode(bl);
+  store->put_bl_ss(bl, COMPAT_SET_LOC, 0);
+}
+
+int Monitor::preinit()
 {
   lock.Lock();
 
-  dout(1) << "init fsid " << monmap->fsid << dendl;
+  dout(1) << "preinit fsid " << monmap->fsid << dendl;
   
   assert(!logger);
   {
@@ -281,22 +461,14 @@ int Monitor::init()
     int r = check_fsid();
     if (r == -ENOENT)
       r = write_fsid();
-    if (r < 0)
+    if (r < 0) {
+      lock.Unlock();
       return r;
+    }
   }
 
   // open compatset
-  {
-    bufferlist bl;
-    store->get_bl_ss(bl, COMPAT_SET_LOC, 0);
-    if (bl.length()) {
-      bufferlist::iterator p = bl.begin();
-      ::decode(features, p);
-    } else {
-      features = get_ceph_mon_feature_compat_set();
-    }
-    dout(10) << "features " << features << dendl;
-  }
+  read_features();
 
   // have we ever joined a quorum?
   has_ever_joined = store->exists_bl_ss("joined");
@@ -330,7 +502,7 @@ int Monitor::init()
   if (authmon()->paxos->get_version() == 0) {
     dout(10) << "loading initial keyring to bootstrap authentication for mkfs" << dendl;
     bufferlist bl;
-    store->get_bl_ss(bl, "mkfs", "keyring");
+    store->get_bl_ss_safe(bl, "mkfs", "keyring");
     KeyRing keyring;
     bufferlist::iterator p = bl.begin();
     ::decode(keyring, p);
@@ -352,6 +524,7 @@ int Monitor::init()
       store->put_bl_ss(bl, "keyring", NULL);
     } else {
       derr << "unable to load initial keyring " << g_conf->keyring << dendl;
+      lock.Unlock();
       return r;
     }
   }
@@ -372,15 +545,24 @@ int Monitor::init()
   assert(r == 0);
   lock.Lock();
 
-  // i'm ready!
-  messenger->add_dispatcher_tail(this);
-  
+  lock.Unlock();
+  return 0;
+}
+
+int Monitor::init()
+{
+  dout(2) << "init" << dendl;
+  lock.Lock();
+
   // start ticker
   timer.init();
   new_tick();
 
+  // i'm ready!
+  messenger->add_dispatcher_tail(this);
+
   bootstrap();
-  
+
   lock.Unlock();
   return 0;
 }
@@ -446,11 +628,16 @@ void Monitor::shutdown()
   for (vector<PaxosService*>::iterator p = paxos_service.begin(); p != paxos_service.end(); p++)
     (*p)->shutdown();
 
+  finish_contexts(g_ceph_context, waitfor_quorum, -ECANCELED);
+  finish_contexts(g_ceph_context, maybe_wait_for_quorum, -ECANCELED);
+
+
   timer.shutdown();
 
   // unlock before msgr shutdown...
   lock.Unlock();
 
+  remove_all_sessions();
   messenger->shutdown();  // last thing!  ceph_mon.cc will delete mon.
 }
 
@@ -549,6 +736,9 @@ void Monitor::reset()
   quorum.clear();
   outside_quorum.clear();
 
+  paxos_recovered.clear();
+  global_version = 0;
+
   for (vector<Paxos*>::iterator p = paxos.begin(); p != paxos.end(); p++)
     (*p)->restart();
   for (vector<PaxosService*>::iterator p = paxos_service.begin(); p != paxos_service.end(); p++)
@@ -847,11 +1037,14 @@ MMonProbe *Monitor::fill_probe_data(MMonProbe *m, Paxos *pax)
   version_t v = MAX(pax->get_first_committed(), m->newest_version + 1);
   int len = 0;
   for (; v <= pax->get_version(); v++) {
-    len += store->get_bl_sn(r->paxos_values[m->machine_name][v], m->machine_name.c_str(), v);
+    store->get_bl_sn_safe(r->paxos_values[m->machine_name][v], m->machine_name.c_str(), v);
+    len += r->paxos_values[m->machine_name][v].length();
+    r->gv[m->machine_name][v] = store->get_global_version(m->machine_name.c_str(), v);
     for (list<string>::iterator p = pax->extra_state_dirs.begin();
          p != pax->extra_state_dirs.end();
          ++p) {
-      len += store->get_bl_sn(r->paxos_values[*p][v], p->c_str(), v);      
+      store->get_bl_sn_safe(r->paxos_values[*p][v], p->c_str(), v);
+      len += r->paxos_values[*p][v].length();
     }
     if (len >= g_conf->mon_slurp_bytes)
       break;
@@ -906,7 +1099,7 @@ void Monitor::handle_probe_data(MMonProbe *m)
     for (map<string, map<version_t, bufferlist> >::iterator p = m->paxos_values.begin();
 	 p != m->paxos_values.end();
 	 ++p) {
-      store->put_bl_sn_map(p->first.c_str(), p->second.begin(), p->second.end());
+      store->put_bl_sn_map(p->first.c_str(), p->second.begin(), p->second.end(), &m->gv[p->first]);
     }
 
     pax->last_committed = m->paxos_values.begin()->second.rbegin()->first;
@@ -962,7 +1155,7 @@ epoch_t Monitor::get_epoch()
   return elector.get_epoch();
 }
 
-void Monitor::win_election(epoch_t epoch, set<int>& active, unsigned features) 
+void Monitor::win_election(epoch_t epoch, set<int>& active, uint64_t features) 
 {
   if (!is_electing())
     reset();
@@ -988,17 +1181,17 @@ void Monitor::win_election(epoch_t epoch, set<int>& active, unsigned features)
   finish_election();
 }
 
-void Monitor::lose_election(epoch_t epoch, set<int> &q, int l) 
+void Monitor::lose_election(epoch_t epoch, set<int> &q, int l, uint64_t features) 
 {
   state = STATE_PEON;
   leader_since = utime_t();
   leader = l;
   quorum = q;
   outside_quorum.clear();
-  quorum_features = 0;
+  quorum_features = features;
   dout(10) << "lose_election, epoch " << epoch << " leader is mon" << leader
-	   << " quorum is " << quorum << dendl;
-  
+	   << " quorum is " << quorum << " features are " << quorum_features << dendl;
+
   for (vector<Paxos*>::iterator p = paxos.begin(); p != paxos.end(); p++)
     (*p)->peon_init();
   for (vector<PaxosService*>::iterator p = paxos_service.begin(); p != paxos_service.end(); p++)
@@ -1155,8 +1348,10 @@ void Monitor::get_health(string& status, bufferlist *detailbl, Formatter *f)
   while (!detail.empty()) {
     if (f)
       f->dump_string("item", detail.front().second);
-    detailbl->append(detail.front().second);
-    detailbl->append('\n');
+    if (detailbl != NULL) {
+      detailbl->append(detail.front().second);
+      detailbl->append('\n');
+    }
     detail.pop_front();
   }
   if (f)
@@ -1329,6 +1524,7 @@ void Monitor::handle_command(MMonCommand *m)
       ss2 << "\n-------- END REPORT " << bl.crc32c(6789) << " --------\n";
       rdata.append(bl);
       rdata.append(ss2.str());
+      rs = string();
       r = 0;
     }
     if (m->cmd[0] == "quorum_status") {
@@ -1511,7 +1707,7 @@ void Monitor::handle_forward(MForward *m)
   m->put();
 }
 
-void Monitor::try_send_message(Message *m, entity_inst_t to)
+void Monitor::try_send_message(Message *m, const entity_inst_t& to)
 {
   dout(10) << "try_send_message " << *m << " to " << to << dendl;
 
@@ -1547,6 +1743,29 @@ void Monitor::send_reply(PaxosServiceMessage *req, Message *reply)
   session->put();
 }
 
+void Monitor::no_reply(PaxosServiceMessage *req)
+{
+  MonSession *session = (MonSession*)req->get_connection()->get_priv();
+  if (!session) {
+    dout(2) << "no_reply no session, dropping non-reply to " << req << " " << *req << dendl;
+    return;
+  }
+  if (session->proxy_con) {
+    if (get_quorum_features() & CEPH_FEATURE_MON_NULLROUTE) {
+      dout(10) << "no_reply to " << req->get_source_inst() << " via mon." << req->session_mon
+	       << " for request " << *req << dendl;
+      messenger->send_message(new MRoute(session->proxy_tid, NULL),
+			      session->proxy_con);
+    } else {
+      dout(10) << "no_reply no quorum nullroute feature for " << req->get_source_inst() << " via mon." << req->session_mon
+	       << " for request " << *req << dendl;
+    }
+  } else {
+    dout(10) << "no_reply to " << req->get_source_inst() << " " << *req << dendl;
+  }
+  session->put();
+}
+
 void Monitor::handle_route(MRoute *m)
 {
   MonSession *session = (MonSession *)m->get_connection()->get_priv();
@@ -1558,7 +1777,10 @@ void Monitor::handle_route(MRoute *m)
     m->put();
     return;
   }
-  dout(10) << "handle_route " << *m->msg << " to " << m->dest << dendl;
+  if (m->msg)
+    dout(10) << "handle_route " << *m->msg << " to " << m->dest << dendl;
+  else
+    dout(10) << "handle_route null to " << m->dest << dendl;
   
   // look it up
   if (m->session_mon_tid) {
@@ -1566,10 +1788,11 @@ void Monitor::handle_route(MRoute *m)
       RoutedRequest *rr = routed_requests[m->session_mon_tid];
 
       // reset payload, in case encoding is dependent on target features
-      m->msg->clear_payload();
-
-      messenger->send_message(m->msg, rr->session->inst);
-      m->msg = NULL;
+      if (m->msg) {
+	m->msg->clear_payload();
+	messenger->send_message(m->msg, rr->session->inst);
+	m->msg = NULL;
+      }
       routed_requests.erase(m->session_mon_tid);
       rr->session->routed_request_tids.insert(rr->tid);
       delete rr;
@@ -1578,8 +1801,10 @@ void Monitor::handle_route(MRoute *m)
     }
   } else {
     dout(10) << " not a routed request, trying to send anyway" << dendl;
-    messenger->lazy_send_message(m->msg, m->dest);
-    m->msg = NULL;
+    if (m->msg) {
+      messenger->lazy_send_message(m->msg, m->dest);
+      m->msg = NULL;
+    }
   }
   m->put();
   if (session)
@@ -1620,9 +1845,17 @@ void Monitor::remove_session(MonSession *s)
       routed_requests.erase(*p);
     }
   }
+  s->con->set_priv(NULL);
   session_map.remove_session(s);
 }
 
+void Monitor::remove_all_sessions()
+{
+  while (!session_map.sessions.empty()) {
+    MonSession *s = session_map.sessions.front();
+    remove_session(s);
+  }
+}
 
 void Monitor::send_command(const entity_inst_t& inst,
 			   const vector<string>& com, version_t version)
@@ -1813,15 +2046,14 @@ bool Monitor::_ms_dispatch(Message *m)
       // paxos
     case MSG_MON_PAXOS:
       {
+	MMonPaxos *pm = (MMonPaxos*)m;
 	if (!src_is_mon && 
 	    !s->caps.check_privileges(PAXOS_MONMAP, MON_CAP_X)) {
 	  //can't send these!
-	  m->put();
+	  pm->put();
 	  break;
 	}
 
-	MMonPaxos *pm = (MMonPaxos*)m;
-
 	// sanitize
 	if (pm->epoch > get_epoch()) {
 	  bootstrap();
@@ -1851,6 +2083,8 @@ bool Monitor::_ms_dispatch(Message *m)
 	  !s->caps.check_privileges(PAXOS_MONMAP, MON_CAP_X)) {
 	dout(0) << "MMonElection received from entity without enough caps!"
 		<< s->caps << dendl;
+	m->put();
+	break;
       }
       if (!is_probing() && !is_slurping()) {
 	elector.dispatch(m);
@@ -1982,9 +2216,6 @@ bool Monitor::ms_handle_reset(Connection *con)
   if (!s->closed)
     remove_session(s);
   s->put();
-    
-  // remove from connection, too.
-  con->set_priv(NULL);
   return true;
 }
 
@@ -2134,7 +2365,8 @@ int Monitor::write_fsid()
 
   bufferlist b;
   b.append(us);
-  return store->put_bl_ss(b, "cluster_uuid", 0);
+  store->put_bl_ss(b, "cluster_uuid", 0);
+  return 0;
 }
 
 /*
@@ -2158,14 +2390,11 @@ int Monitor::mkfs(bufferlist& osdmapbl)
   bufferlist magicbl;
   magicbl.append(CEPH_MON_ONDISK_MAGIC);
   magicbl.append("\n");
-  r = store->put_bl_ss(magicbl, "magic", 0);
-  if (r < 0)
-    return r;
+  store->put_bl_ss(magicbl, "magic", 0);
 
-  bufferlist features;
-  CompatSet mon_features = get_ceph_mon_feature_compat_set();
-  mon_features.encode(features);
-  store->put_bl_ss(features, COMPAT_SET_LOC, 0);
+
+  features = get_supported_features();
+  write_features();
 
   // save monmap, osdmap, keyring.
   bufferlist monmapbl;
@@ -2187,7 +2416,13 @@ int Monitor::mkfs(bufferlist& osdmapbl)
   }
 
   KeyRing keyring;
-  r = keyring.load(g_ceph_context, g_conf->keyring);
+  string keyring_filename;
+  if (!ceph_resolve_file_search(g_conf->keyring, keyring_filename)) {
+    derr << "unable to find a keyring file on " << g_conf->keyring << dendl;
+    return -ENOENT;
+  }
+
+  r = keyring.load(g_ceph_context, keyring_filename);
   if (r < 0) {
     derr << "unable to load initial keyring " << g_conf->keyring << dendl;
     return r;
@@ -2287,7 +2522,7 @@ bool Monitor::ms_get_authorizer(int service_id, AuthAuthorizer **authorizer, boo
 
 bool Monitor::ms_verify_authorizer(Connection *con, int peer_type,
 				   int protocol, bufferlist& authorizer_data, bufferlist& authorizer_reply,
-				   bool& isvalid)
+				   bool& isvalid, CryptoKey& session_key)
 {
   dout(10) << "ms_verify_authorizer " << con->get_peer_addr()
 	   << " " << ceph_entity_type_name(peer_type)
@@ -2307,10 +2542,12 @@ bool Monitor::ms_verify_authorizer(Connection *con, int peer_type,
       if (authorizer_data.length()) {
 	int ret = cephx_verify_authorizer(g_ceph_context, &keyring, iter,
 					  auth_ticket_info, authorizer_reply);
-	if (ret >= 0)
+	if (ret >= 0) {
+	  session_key = auth_ticket_info.session_key;
 	  isvalid = true;
-	else
+	} else {
 	  dout(0) << "ms_verify_authorizer bad authorizer from mon " << con->get_peer_addr() << dendl;
+        }
       }
     } else {
       dout(0) << "ms_verify_authorizer cephx enabled, but no authorizer (required for mon)" << dendl;
