@@ -75,7 +75,7 @@ void OSDMonitor::create_initial()
   OSDMap newmap;
 
   bufferlist bl;
-  mon->store->get_bl_ss(bl, "mkfs", "osdmap");
+  mon->store->get_bl_ss_safe(bl, "mkfs", "osdmap");
   if (bl.length()) {
     newmap.decode(bl);
     newmap.set_fsid(mon->monmap->fsid);
@@ -151,6 +151,8 @@ void OSDMonitor::update_from_paxos()
   share_map_with_random_osd();
   update_logger();
 
+  process_failures();
+
   // make sure our feature bits reflect the latest map
   update_msgr_features();
 }
@@ -163,23 +165,19 @@ void OSDMonitor::update_msgr_features()
   types.insert((int)entity_name_t::TYPE_MDS);
   types.insert((int)entity_name_t::TYPE_MON);
 
-  if (osdmap.crush->has_nondefault_tunables()) {
-    for (set<int>::iterator q = types.begin(); q != types.end(); ++q) {
-      if (!(mon->messenger->get_policy(*q).features_required & CEPH_FEATURE_CRUSH_TUNABLES)) {
-	dout(0) << "crush map has non-default tunables, requiring CRUSH_TUNABLES feature" << dendl;
-	Messenger::Policy p = mon->messenger->get_policy(*q);
-	p.features_required |= CEPH_FEATURE_CRUSH_TUNABLES;
-	mon->messenger->set_policy(*q, p);
-      }
-    }
-  } else {
-    for (set<int>::iterator q = types.begin(); q != types.end(); ++q) {
-      if (mon->messenger->get_policy(*q).features_required & CEPH_FEATURE_CRUSH_TUNABLES) {
-	dout(0) << "crush map has default tunables, not requiring CRUSH_TUNABLES feature" << dendl;
-	Messenger::Policy p = mon->messenger->get_policy(*q);
-	p.features_required &= ~CEPH_FEATURE_CRUSH_TUNABLES;
-	mon->messenger->set_policy(*q, p);
-      }
+  uint64_t mask = CEPH_FEATURES_CRUSH;
+  uint64_t features = 0;
+  if (osdmap.crush->has_nondefault_tunables())
+    features |= CEPH_FEATURE_CRUSH_TUNABLES;
+  if (osdmap.crush->has_nondefault_tunables2())
+    features |= CEPH_FEATURE_CRUSH_TUNABLES2;
+
+  for (set<int>::iterator q = types.begin(); q != types.end(); ++q) {
+    if ((mon->messenger->get_policy(*q).features_required & mask) != features) {
+      dout(0) << "crush map has features " << features << ", adjusting msgr requires" << dendl;
+      Messenger::Policy p = mon->messenger->get_policy(*q);
+      p.features_required = (p.features_required & ~mask) | features;
+      mon->messenger->set_policy(*q, p);
     }
   }
 }
@@ -270,6 +268,10 @@ void OSDMonitor::on_active()
 
   if (mon->is_leader())
     mon->clog.info() << "osdmap " << osdmap << "\n"; 
+
+  if (!mon->is_leader()) {
+    kick_all_failures();
+  }
 }
 
 void OSDMonitor::update_logger()
@@ -727,6 +729,12 @@ bool OSDMonitor::check_failure(utime_t now, int target_osd, failure_info_t& fi)
 	   << grace << " grace (" << orig_grace << " + " << my_grace << " + " << peer_grace << "), max_failed_since " << max_failed_since
 	   << dendl;
 
+  // already pending failure?
+  if (pending_inc.new_state[target_osd] & CEPH_OSD_UP) {
+    dout(10) << " already pending failure" << dendl;
+    return true;
+  }
+
   if (failed_for >= grace &&
       ((int)fi.reporters.size() >= g_conf->osd_min_down_reporters) &&
       (fi.num_reports >= g_conf->osd_min_down_reports)) {
@@ -736,12 +744,6 @@ bool OSDMonitor::check_failure(utime_t now, int target_osd, failure_info_t& fi)
     mon->clog.info() << osdmap.get_inst(target_osd) << " failed ("
 		     << fi.num_reports << " reports from " << (int)fi.reporters.size() << " peers after "
 		     << failed_for << " >= grace " << grace << ")\n";
-
-    list<MOSDFailure*> msgs;
-    fi.take_report_messages(msgs);
-    failure_info.erase(target_osd);
-
-    paxos->wait_for_commit(new C_Reported(this, msgs));
     return true;
   }
   return false;
@@ -762,7 +764,7 @@ bool OSDMonitor::prepare_failure(MOSDFailure *m)
 
   // calculate failure time
   utime_t now = ceph_clock_now(g_ceph_context);
-  utime_t failed_since = now - utime_t(m->failed_for ? m->failed_for : g_conf->osd_heartbeat_grace, 0);
+  utime_t failed_since = m->get_recv_stamp() - utime_t(m->failed_for ? m->failed_for : g_conf->osd_heartbeat_grace, 0);
   
   if (m->if_osd_failed()) {
     // add a report
@@ -794,13 +796,42 @@ bool OSDMonitor::prepare_failure(MOSDFailure *m)
   return false;
 }
 
-void OSDMonitor::_reported_failure(list<MOSDFailure*>& ls)
+void OSDMonitor::process_failures()
 {
-  dout(7) << "_reported_failure telling " << ls << dendl;
+  map<int,failure_info_t>::iterator p = failure_info.begin();
+  while (p != failure_info.end()) {
+    if (osdmap.is_up(p->first)) {
+      ++p;
+    } else {
+      dout(10) << "process_failures osd." << p->first << dendl;
+      list<MOSDFailure*> ls;
+      p->second.take_report_messages(ls);
+      failure_info.erase(p++);
+
+      while (!ls.empty()) {
+	send_latest(ls.front(), ls.front()->get_epoch());
+	ls.pop_front();
+      }
+    }
+  }
+}
+
+void OSDMonitor::kick_all_failures()
+{
+  dout(10) << "kick_all_failures on " << failure_info.size() << " osds" << dendl;
+  assert(!mon->is_leader());
+
+  list<MOSDFailure*> ls;
+  for (map<int,failure_info_t>::iterator p = failure_info.begin();
+       p != failure_info.end();
+       ++p) {
+    p->second.take_report_messages(ls);
+  }
+  failure_info.clear();
+
   while (!ls.empty()) {
-    MOSDFailure *m = ls.front();
+    dispatch(ls.front());
     ls.pop_front();
-    send_latest(m, m->get_epoch());
   }
 }
 
@@ -1278,7 +1309,8 @@ void OSDMonitor::send_incremental(PaxosServiceMessage *req, epoch_t first)
   if (first < paxos->get_first_committed()) {
     first = paxos->get_first_committed();
     bufferlist bl;
-    mon->store->get_bl_sn(bl, "osdmap_full", first);
+    mon->store->get_bl_sn_safe(bl, "osdmap_full", first);
+    assert(bl.length());
     dout(20) << "send_incremental starting with base full " << first << " " << bl.length() << " bytes" << dendl;
     MOSDMap *m = new MOSDMap(osdmap.get_fsid());
     m->oldest_map = paxos->get_first_committed();
@@ -1305,7 +1337,8 @@ void OSDMonitor::send_incremental(epoch_t first, entity_inst_t& dest, bool oneti
   if (first < paxos->get_first_committed()) {
     first = paxos->get_first_committed();
     bufferlist bl;
-    mon->store->get_bl_sn(bl, "osdmap_full", first);
+    mon->store->get_bl_sn_safe(bl, "osdmap_full", first);
+    assert(bl.length());
     dout(20) << "send_incremental starting with base full " << first << " " << bl.length() << " bytes" << dendl;
     MOSDMap *m = new MOSDMap(osdmap.get_fsid());
     m->oldest_map = paxos->get_first_committed();
@@ -1328,7 +1361,7 @@ void OSDMonitor::send_incremental(epoch_t first, entity_inst_t& dest, bool oneti
 
 
 
-epoch_t OSDMonitor::blacklist(entity_addr_t a, utime_t until)
+epoch_t OSDMonitor::blacklist(const entity_addr_t& a, utime_t until)
 {
   dout(10) << "blacklist " << a << " until " << until << dendl;
   pending_inc.new_blacklist[a] = until;
@@ -1589,6 +1622,22 @@ void OSDMonitor::get_health(list<pair<health_status_t,string> >& summary,
 	}
       }
     }
+
+    // warn about flags
+    if (osdmap.test_flag(CEPH_OSDMAP_PAUSERD |
+			 CEPH_OSDMAP_PAUSEWR |
+			 CEPH_OSDMAP_NOUP |
+			 CEPH_OSDMAP_NODOWN |
+			 CEPH_OSDMAP_NOIN |
+			 CEPH_OSDMAP_NOOUT |
+			 CEPH_OSDMAP_NOBACKFILL |
+			 CEPH_OSDMAP_NORECOVER)) {
+      ostringstream ss;
+      ss << osdmap.get_flag_string() << " flag(s) set";
+      summary.push_back(make_pair(HEALTH_WARN, ss.str()));
+      if (detail)
+	detail->push_back(make_pair(HEALTH_WARN, ss.str()));
+    }
   }
 }
 
@@ -1597,6 +1646,10 @@ void OSDMonitor::dump_info(Formatter *f)
   f->open_object_section("osdmap");
   osdmap.dump(f);
   f->close_section();
+
+  f->open_object_section("crushmap");
+  osdmap.crush->dump(f);
+  f->close_section();
 }
 
 bool OSDMonitor::preprocess_command(MMonCommand *m)
@@ -1650,7 +1703,7 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
       OSDMap *p = &osdmap;
       if (epoch) {
 	bufferlist b;
-	mon->store->get_bl_sn(b,"osdmap_full", epoch);
+	mon->store->get_bl_sn_safe(b,"osdmap_full", epoch);
 	if (!b.length()) {
 	  p = 0;
 	  r = -ENOENT;
@@ -1678,10 +1731,23 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
 	  }
 	} else if (cmd == "tree") {
 	  stringstream ds;
-	  p->print_tree(ds);
-	  rdata.append(ds);
-	  ss << "dumped osdmap tree epoch " << p->get_epoch();
-	  r = 0;
+	  if (format == "json") {
+	    JSONFormatter jf(true);
+	    jf.open_object_section("tree");
+	    p->print_tree(NULL, &jf);
+	    jf.close_section();
+	    jf.flush(ds);
+	    r = 0;
+	  } else if (format == "plain") {
+	    p->print_tree(&ds, NULL);
+	    r = 0;
+	  } else {
+	    ss << "unrecognized format '" << format << "'";
+	    r = -EINVAL;
+	  }
+	  if (r == 0) {
+	    rdata.append(ds);
+	  }
 	} else if (cmd == "getmap") {
 	  p->encode(rdata);
 	  ss << "got osdmap epoch " << p->get_epoch();
@@ -1716,9 +1782,11 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
 	ss << "ok";
       } else {
 	errno = 0;
-	int who = strtol(m->cmd[0].c_str(), 0, 10);
+	int who = parse_osd_id(m->cmd[0].c_str(), &ss);
 	m->cmd.erase(m->cmd.begin()); //done with target num now
-	if (!errno && who >= 0) {
+	if (who < 0) {
+	  r = -EINVAL;
+	} else {
 	  if (osdmap.is_up(who)) {
 	    mon->send_command(osdmap.get_inst(who), m->cmd, paxos->get_version());
 	    r = 0;
@@ -1727,13 +1795,13 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
 	    ss << "osd." << who << " not up";
 	    r = -ENOENT;
 	  }
-	} else ss << "specify osd number or *";
+	}
       }
     }
     else if (m->cmd[1] == "map" && m->cmd.size() == 4) {
       int64_t pool = osdmap.lookup_pg_pool_name(m->cmd[2].c_str());
       if (pool < 0) {
-	ss << "pool " << m->cmd[2] << " dne";
+	ss << "pool " << m->cmd[2] << " does not exist";
 	r = -ENOENT;
       } else {
 	object_locator_t oloc(pool);
@@ -1771,16 +1839,19 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
 	r = 0;
 	ss << " instructed to " << m->cmd[1];
       } else {
-	long osd = strtol(m->cmd[2].c_str(), 0, 10);
-	if (osdmap.is_up(osd)) {
+	long osd = parse_osd_id(m->cmd[2].c_str(), &ss);
+	if (osd < 0) {
+	  r = -EINVAL;
+	} else if (osdmap.is_up(osd)) {
 	  mon->try_send_message(new MOSDScrub(osdmap.get_fsid(),
 					      m->cmd[1] == "repair",
 					      m->cmd[1] == "deep-scrub"),
 				osdmap.get_inst(osd));
 	  r = 0;
 	  ss << "osd." << osd << " instructed to " << m->cmd[1];
-	} else 
+	} else {
 	  ss << "osd." << osd << " is not up";
+	}
       }
     }
     else if (m->cmd[1] == "lspools") {
@@ -1859,6 +1930,7 @@ int OSDMonitor::prepare_new_pool(string& name, uint64_t auid, int crush_rule,
   pending_inc.new_pools[pool].type = pg_pool_t::TYPE_REP;
 
   pending_inc.new_pools[pool].size = g_conf->osd_pool_default_size;
+  pending_inc.new_pools[pool].min_size = g_conf->get_osd_pool_default_min_size();
   if (crush_rule >= 0)
     pending_inc.new_pools[pool].crush_ruleset = crush_rule;
   else
@@ -1963,13 +2035,6 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	goto out;
       }
 
-      if (crush.get_max_devices() > osdmap.get_max_osd()) {
-	err = -ERANGE;
-	ss << "crushmap max_devices " << crush.get_max_devices()
-	   << " > osdmap max_osd " << osdmap.get_max_osd();
-	goto out;
-      }
-
       // sanity check: test some inputs to make sure this map isn't totally broken
       dout(10) << " testing map" << dendl;
       stringstream ess;
@@ -2108,7 +2173,7 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 
 	if (!newcrush.name_exists(name.c_str())) {
 	  err = -ENOENT;
-	  ss << "item " << name << " dne";
+	  ss << "item " << name << " does not exist";
 	  break;
 	}
 	int id = newcrush.get_item_id(name.c_str());
@@ -2141,9 +2206,14 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	bufferlist::iterator p = bl.begin();
 	newcrush.decode(p);
 
+	if (!newcrush.name_exists(m->cmd[3].c_str())) {
+	  err = -ENOENT;
+	  ss << "device '" << m->cmd[3] << "' does not appear in the crush map";
+	  break;
+	}
 	int id = newcrush.get_item_id(m->cmd[3].c_str());
 	if (id < 0) {
-	  ss << "device '" << m->cmd[3] << "' does not appear in the crush map";
+	  ss << "item '" << m->cmd[3] << "' is not a leaf in the crush map";
 	  break;
 	}
 	err = newcrush.remove_item(g_ceph_context, id);
@@ -2170,9 +2240,15 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	bufferlist::iterator p = bl.begin();
 	newcrush.decode(p);
 
+	if (!newcrush.name_exists(m->cmd[3].c_str())) {
+	  err = -ENOENT;
+	  ss << "device '" << m->cmd[3] << "' does not appear in the crush map";
+	  break;
+	}
+
 	int id = newcrush.get_item_id(m->cmd[3].c_str());
 	if (id < 0) {
-	  ss << "device '" << m->cmd[3] << "' does not appear in the crush map";
+	  ss << "device '" << m->cmd[3] << "' is not a leaf in the crush map";
 	  break;
 	}
 	float w = atof(m->cmd[4].c_str());
@@ -2195,12 +2271,6 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	err = -EINVAL;
 	goto out;
       }
-      if (newmax < osdmap.crush->get_max_devices()) {
-	err = -ERANGE;
-	ss << "cannot set max_osd to " << newmax << " which is < crush max_devices "
-	   << osdmap.crush->get_max_devices();
-	goto out;
-      }
       if (newmax > g_conf->mon_max_osd) {
 	err = -ERANGE;
 	ss << "cannot set max_osd to " << newmax << " which is > conf.mon_max_osd ("
@@ -2270,18 +2340,17 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
     else if (m->cmd[1] == "down" && m->cmd.size() >= 3) {
       bool any = false;
       for (unsigned j = 2; j < m->cmd.size(); j++) {
-	long osd = strtol(m->cmd[j].c_str(), 0, 10);
-	if (!osdmap.exists(osd)) {
-	  ss << "osd." << osd << " does not exist";
+	long osd = parse_osd_id(m->cmd[j].c_str(), &ss);
+	if (osd < 0) {
+	  err = -EINVAL;
+	} else if (!osdmap.exists(osd)) {
+	  ss << "osd." << osd << " does not exist. ";
 	} else if (osdmap.is_down(osd)) {
-	  ss << "osd." << osd << " is already down";
+	  ss << "osd." << osd << " is already down. ";
 	  err = 0;
 	} else {
 	  pending_inc.new_state[osd] = CEPH_OSD_UP;
-	  if (any)
-	    ss << ", osd." << osd;
-	  else 
-	    ss << "marked down osd." << osd;
+	  ss << "marked down osd." << osd << ". ";
 	  any = true;
 	}
       }
@@ -2294,18 +2363,17 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
     else if (m->cmd[1] == "out" && m->cmd.size() >= 3) {
       bool any = false;
       for (unsigned j = 2; j < m->cmd.size(); j++) {
-	long osd = strtol(m->cmd[j].c_str(), 0, 10);
-	if (!osdmap.exists(osd)) {
-	  ss << "osd." << osd << " does not exist";
+	long osd = parse_osd_id(m->cmd[j].c_str(), &ss);
+	if (osd < 0) {
+	  err = -EINVAL;
+	} else if (!osdmap.exists(osd)) {
+	  ss << "osd." << osd << " does not exist. ";
 	} else if (osdmap.is_out(osd)) {
-	  ss << "osd." << osd << " is already out";
+	  ss << "osd." << osd << " is already out. ";
 	  err = 0;
 	} else {
 	  pending_inc.new_weight[osd] = CEPH_OSD_OUT;
-	  if (any)
-	    ss << ", osd." << osd;
-	  else
-	    ss << "marked out osd." << osd;
+	  ss << "marked out osd." << osd << ". ";
 	  any = true;
 	}
       }
@@ -2318,18 +2386,17 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
     else if (m->cmd[1] == "in" && m->cmd.size() >= 3) {
       bool any = false;
       for (unsigned j = 2; j < m->cmd.size(); j++) {
-	long osd = strtol(m->cmd[j].c_str(), 0, 10);
-	if (osdmap.is_in(osd)) {
-	  ss << "osd." << osd << " is already in";
+	long osd = parse_osd_id(m->cmd[j].c_str(), &ss);
+	if (osd < 0) {
+	  err = -EINVAL;
+	} else if (osdmap.is_in(osd)) {
+	  ss << "osd." << osd << " is already in. ";
 	  err = 0;
 	} else if (!osdmap.exists(osd)) {
-	  ss << "osd." << osd << " does not exist";
+	  ss << "osd." << osd << " does not exist. ";
 	} else {
 	  pending_inc.new_weight[osd] = CEPH_OSD_IN;
-	  if (any)
-	    ss << ", osd." << osd;
-	  else
-	    ss << "marked in osd." << osd;
+	  ss << "marked in osd." << osd << ". ";
 	  any = true;
 	}
       }
@@ -2340,22 +2407,25 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
       } 
     }
     else if (m->cmd[1] == "reweight" && m->cmd.size() == 4) {
-      long osd = strtol(m->cmd[2].c_str(), 0, 10);
-      float w = strtof(m->cmd[3].c_str(), 0);
-      long ww = (int)((float)CEPH_OSD_IN*w);
-      if (osdmap.exists(osd)) {
-	pending_inc.new_weight[osd] = ww;
-	ss << "reweighted osd." << osd << " to " << w << " (" << ios::hex << ww << ios::dec << ")";
-	getline(ss, rs);
-	paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
-	return true;
-      } 
+      long osd = parse_osd_id(m->cmd[2].c_str(), &ss);
+      if (osd < 0) {
+	err = -EINVAL;
+      } else {
+	float w = strtof(m->cmd[3].c_str(), 0);
+	long ww = (int)((float)CEPH_OSD_IN*w);
+	if (osdmap.exists(osd)) {
+	  pending_inc.new_weight[osd] = ww;
+	  ss << "reweighted osd." << osd << " to " << w << " (" << ios::hex << ww << ios::dec << ")";
+	  getline(ss, rs);
+	  paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
+	  return true;
+	}
+      }
     }
     else if (m->cmd[1] == "lost" && m->cmd.size() >= 3) {
-      string err;
-      int osd = strict_strtol(m->cmd[2].c_str(), 10, &err);
-      if (!err.empty()) {
-	ss << err;
+      int osd = parse_osd_id(m->cmd[2].c_str(), &ss);
+      if (osd < 0) {
+	err = -EINVAL;
       }
       else if ((m->cmd.size() < 4) || m->cmd[3] != "--yes-i-really-mean-it") {
 	ss << "are you SURE?  this might mean real, permanent data loss.  pass "
@@ -2377,7 +2447,11 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 
       // optional uuid provided?
       uuid_d uuid;
-      if (m->cmd.size() > 2 && uuid.parse(m->cmd[2].c_str())) {
+      if (m->cmd.size() > 2) {
+        if (!uuid.parse(m->cmd[2].c_str())) {
+          err = -EINVAL;
+          goto out;
+        }
 	dout(10) << " osd create got uuid " << uuid << dendl;
 	i = osdmap.identify_osd(uuid);
 	if (i >= 0) {
@@ -2424,18 +2498,20 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
     else if (m->cmd[1] == "rm" && m->cmd.size() >= 3) {
       bool any = false;
       for (unsigned j = 2; j < m->cmd.size(); j++) {
-	long osd = strtol(m->cmd[j].c_str(), 0, 10);
-	if (!osdmap.exists(osd)) {
-	  ss << "osd." << osd << " does not exist";
+	long osd = parse_osd_id(m->cmd[j].c_str(), &ss);
+	if (osd < 0) {
+	  err = -EINVAL;
+	} else if (!osdmap.exists(osd)) {
+	  ss << "osd." << osd << " does not exist. ";
 	  err = 0;
 	} else if (osdmap.is_up(osd)) {
-	  ss << "osd." << osd << " is still up";
+	  ss << "osd." << osd << " is still up; must be down before removal. ";
 	} else {
 	  pending_inc.new_state[osd] = osdmap.get_state(osd);
 	  if (any)
 	    ss << ", osd." << osd;
 	  else 
-	    ss << "marked dne osd." << osd;
+	    ss << "removed osd." << osd;
 	  any = true;
 	}
       }
@@ -2539,27 +2615,30 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	  }
 	}
       }
-      else if (m->cmd[2] == "create" && m->cmd.size() >= 4) {
+      else if (m->cmd[2] == "create" && m->cmd.size() >= 3) {
+	if (m->cmd.size() < 5) {
+	  ss << "usage: osd pool create <poolname> <pg_num> [pgp_num]";
+	  err = -EINVAL;
+	  goto out;
+	}
         int pg_num = 0;
         int pgp_num = 0;
-        if (m->cmd.size() > 4) { // try to parse out pg_num and pgp_num
-          const char *start = m->cmd[4].c_str();
-          char *end = (char*)start;
-          pgp_num = pg_num = strtol(start, &end, 10);
-          if (*end != '\0') { // failed to parse
-            err = -EINVAL;
-            ss << "usage: osd pool create <poolname> [pg_num [pgp_num]]";
-            goto out;
-          } else if (m->cmd.size() > 5) { // check for pgp_num too
-            start = m->cmd[5].c_str();
-            end = (char *)start;
-            pgp_num = strtol(start, &end, 10);
-            if (*end != '\0') { // failed to parse
-              err = -EINVAL;
-              ss << "usage: osd pool create <poolname> [pg_num [pgp_num]]";
-              goto out;
-            }
-          }
+	const char *start = m->cmd[4].c_str();
+	char *end = (char*)start;
+	pgp_num = pg_num = strtol(start, &end, 10);
+	if (*end != '\0') { // failed to parse
+	  err = -EINVAL;
+	  ss << "usage: osd pool create <poolname> <pg_num> [pgp_num]";
+	  goto out;
+	} else if (m->cmd.size() > 5) { // check for pgp_num too
+	  start = m->cmd[5].c_str();
+	  end = (char *)start;
+	  pgp_num = strtol(start, &end, 10);
+	  if (*end != '\0') { // failed to parse
+	    err = -EINVAL;
+	    ss << "usage: osd pool create <poolname> <pg_num> [pgp_num]";
+	    goto out;
+	  }
         }
 
 	if (osdmap.name_pool.count(m->cmd[3])) {
@@ -2636,11 +2715,22 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	      if (pending_inc.new_pools.count(pool) == 0)
 		pending_inc.new_pools[pool] = *p;
 	      pending_inc.new_pools[pool].size = n;
+	      if (n < p->min_size)
+		pending_inc.new_pools[pool].min_size = n;
 	      pending_inc.new_pools[pool].last_change = pending_inc.epoch;
 	      ss << "set pool " << pool << " size to " << n;
 	      getline(ss, rs);
 	      paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
 	      return true;
+	    } else if (m->cmd[4] == "min_size") {
+	      if (pending_inc.new_pools.count(pool) == 0)
+		pending_inc.new_pools[pool] = *p;
+	      pending_inc.new_pools[pool].min_size = n;
+	      pending_inc.new_pools[pool].last_change = pending_inc.epoch;
+	      ss << "set pool " << pool << " min_size to " << n;
+	      getline(ss, rs);
+	      paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
+	      return true;
 	    } else if (m->cmd[4] == "crash_replay_interval") {
 	      if (pending_inc.new_pools.count(pool) == 0)
 		pending_inc.new_pools[pool] = *p;
@@ -2697,7 +2787,7 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 		paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
 		return true;
 	      } else {
-		ss << "crush ruleset " << n << " dne";
+		ss << "crush ruleset " << n << " does not exist";
 		err = -ENOENT;
 	      }
 	    } else {
