@@ -31,11 +31,11 @@ using namespace std;
 #include "Session.h"
 
 class Monitor;
-class MOSDBoot;
-class MMonCommand;
-class MPoolSnap;
-class MOSDMap;
-class MOSDFailure;
+#include "messages/MOSDBoot.h"
+#include "messages/MMonCommand.h"
+#include "messages/MOSDMap.h"
+#include "messages/MOSDFailure.h"
+#include "messages/MPoolOp.h"
 
 /// information about a particular peer's failure reports for one osd
 struct failure_reporter_t {
@@ -169,7 +169,8 @@ private:
  
   bool preprocess_failure(class MOSDFailure *m);
   bool prepare_failure(class MOSDFailure *m);
-  void _reported_failure(list<MOSDFailure*>& m);
+  void process_failures();
+  void kick_all_failures();
 
   bool preprocess_boot(class MOSDBoot *m);
   bool prepare_boot(class MOSDBoot *m);
@@ -207,6 +208,12 @@ private:
     C_Booted(OSDMonitor *cm, MOSDBoot *m_, bool l=true) : 
       cmon(cm), m(m_), logit(l) {}
     void finish(int r) {
+      if (r == -ECANCELED) {
+	if (m)
+	  m->put();
+	return;
+      }
+
       if (r >= 0)
 	cmon->_booted(m, logit);
       else
@@ -220,27 +227,14 @@ private:
     epoch_t e;
     C_ReplyMap(OSDMonitor *o, PaxosServiceMessage *mm, epoch_t ee) : osdmon(o), m(mm), e(ee) {}
     void finish(int r) {
+      if (r == -ECANCELED) {
+	if (m)
+	  m->put();
+	return;
+      }
       osdmon->_reply_map(m, e);
     }    
   };
-  struct C_Reported : public Context {
-    OSDMonitor *cmon;
-    list<MOSDFailure*> msgs;
-    C_Reported(OSDMonitor *cm, list<MOSDFailure*>& m_)
-      : cmon(cm) {
-      msgs.swap(m_);
-    }
-    void finish(int r) {
-      if (r >= 0)
-	cmon->_reported_failure(msgs);
-      else {
-	while (!msgs.empty()) {
-	  cmon->dispatch((PaxosServiceMessage*)msgs.front());
-	  msgs.pop_front();
-	}
-      }
-    }
-  };
   struct C_PoolOp : public Context {
     OSDMonitor *osdmon;
     MPoolOp *m;
@@ -250,6 +244,11 @@ private:
     C_PoolOp(OSDMonitor * osd, MPoolOp *m_, int rc, int e, bufferlist *rd=NULL) : 
       osdmon(osd), m(m_), replyCode(rc), epoch(e), reply_data(rd) {}
     void finish(int r) {
+      if (r == -ECANCELED) {
+	if (m)
+	  m->put();
+	return;
+      }
       osdmon->_pool_op_reply(m, replyCode, epoch, reply_data);
     }
   };
@@ -279,7 +278,7 @@ private:
     send_incremental(m, start);
   }
 
-  epoch_t blacklist(entity_addr_t a, utime_t until);
+  epoch_t blacklist(const entity_addr_t& a, utime_t until);
 
   void dump_info(Formatter *f);
 
