@@ -94,6 +94,7 @@
 #define MSG_COMMAND_REPLY      98
 
 #define MSG_OSD_BACKFILL_RESERVE 99
+#define MSG_OSD_RECOVERY_RESERVE 150
 
 // *** MDS ***
 
@@ -101,6 +102,8 @@
 #define MSG_MDS_SLAVE_REQUEST      101
 #define MSG_MDS_TABLE_REQUEST      102
 
+                                // 150 already in use (MSG_OSD_RECOVERY_RESERVE)
+
 #define MSG_MDS_RESOLVE            0x200
 #define MSG_MDS_RESOLVEACK         0x201
 #define MSG_MDS_CACHEREJOIN        0x202
@@ -477,8 +480,4 @@ inline ostream& operator<<(ostream& out, Message& m) {
 extern void encode_message(Message *m, uint64_t features, bufferlist& bl);
 extern Message *decode_message(CephContext *cct, bufferlist::iterator& bl);
 
-void intrusive_ptr_add_ref(Message *p);
-void intrusive_ptr_release(Message *p);
-
-
 #endif
