@@ -17,6 +17,8 @@
 
 #include "msg_types.h"
 #include "Messenger.h"
+#include "auth/AuthSessionHandler.h"
+
 
 class SimpleMessenger;
 class IncomingQueue;
@@ -63,6 +65,7 @@ class DispatchQueue;
     ~Pipe();
 
     SimpleMessenger *msgr;
+    uint64_t conn_id;
     ostream& _pipe_prefix(std::ostream *_dout);
 
     enum {
@@ -100,17 +103,21 @@ class DispatchQueue;
     Mutex pipe_lock;
     int state;
 
+    // session_security handles any signatures or encryptions required for this pipe's msgs. PLR
+
+    AuthSessionHandler *session_security;
+
   protected:
     friend class SimpleMessenger;
     Connection *connection_state;
 
     utime_t backoff;         // backoff time
 
-    bool reader_running, reader_joining;
+    bool reader_running, reader_needs_join;
     bool writer_running;
 
     map<int, list<Message*> > out_q;  // priority queue for outbound msgs
-    IncomingQueue *in_q;
+    DispatchQueue *in_q;
     list<Message*> sent;
     Cond cond;
     bool keepalive;
@@ -127,6 +134,8 @@ class DispatchQueue;
     void writer();
     void unlock_maybe_reap();
 
+    int randomize_out_seq();
+
     int read_message(Message **pm);
     int write_message(Message *m);
     /**
@@ -144,7 +153,7 @@ class DispatchQueue;
     int write_ack(uint64_t s);
     int write_keepalive();
 
-    void fault(bool onconnect=false, bool reader=false);
+    void fault(bool reader=false);
 
     void was_session_reset();
 
