@@ -1247,8 +1247,13 @@ void OSDMap::dump(Formatter *f) const
 
   f->open_array_section("pools");
   for (map<int64_t,pg_pool_t>::const_iterator p = pools.begin(); p != pools.end(); ++p) {
+    std::string name("<unknown>");
+    map<int64_t,string>::const_iterator pni = pool_name.find(p->first);
+    if (pni != pool_name.end())
+      name = pni->second;
     f->open_object_section("pool");
     f->dump_int("pool", p->first);
+    f->dump_string("pool_name", name);
     p->second.dump(f);
     f->close_section();
   }
@@ -1337,17 +1342,17 @@ string OSDMap::get_flag_string(unsigned f)
   if (f & CEPH_OSDMAP_PAUSEREC)
     s += ",pauserec";
   if (f & CEPH_OSDMAP_NOUP)
-    s += ",no-up";
+    s += ",noup";
   if (f & CEPH_OSDMAP_NODOWN)
-    s += ",no-down";
+    s += ",nodown";
   if (f & CEPH_OSDMAP_NOOUT)
-    s += ",no-out";
+    s += ",noout";
   if (f & CEPH_OSDMAP_NOIN)
-    s += ",no-in";
+    s += ",noin";
   if (f & CEPH_OSDMAP_NOBACKFILL)
-    s += ",no-backfill";
+    s += ",nobackfill";
   if (f & CEPH_OSDMAP_NORECOVER)
-    s += ",no-recover";
+    s += ",norecover";
   if (s.length())
     s = s.erase(0, 1);
   return s;
@@ -1609,6 +1614,7 @@ void OSDMap::build_simple(CephContext *cct, epoch_t e, uuid_d &fsid,
     int64_t pool = ++pool_max;
     pools[pool].type = pg_pool_t::TYPE_REP;
     pools[pool].size = cct->_conf->osd_pool_default_size;
+    pools[pool].min_size = cct->_conf->get_osd_pool_default_min_size();
     pools[pool].crush_ruleset = p->first;
     pools[pool].object_hash = CEPH_STR_HASH_RJENKINS;
     pools[pool].set_pg_num(poolbase << pg_bits);
@@ -1730,6 +1736,7 @@ int OSDMap::build_simple_from_conf(CephContext *cct, epoch_t e, uuid_d &fsid,
     int64_t pool = ++pool_max;
     pools[pool].type = pg_pool_t::TYPE_REP;
     pools[pool].size = cct->_conf->osd_pool_default_size;
+    pools[pool].min_size = cct->_conf->get_osd_pool_default_min_size();
     pools[pool].crush_ruleset = p->first;
     pools[pool].object_hash = CEPH_STR_HASH_RJENKINS;
     pools[pool].set_pg_num((numosd + 1) << pg_bits);
