@@ -383,7 +383,7 @@ public:
     OpContext(OpRequestRef _op, osd_reqid_t _reqid, vector<OSDOp>& _ops,
 	      ObjectState *_obs, SnapSetContext *_ssc,
 	      ReplicatedPG *_pg) :
-      op(_op), reqid(_reqid), ops(_ops), obs(_obs),
+      op(_op), reqid(_reqid), ops(_ops), obs(_obs), snapset(0),
       new_obs(_obs->oi, _obs->exists),
       modify(false), user_modify(false),
       watch_connect(false), watch_disconnect(false),
@@ -576,6 +576,7 @@ protected:
   struct PushInfo {
     ObjectRecoveryProgress recovery_progress;
     ObjectRecoveryInfo recovery_info;
+    int priority;
 
     void dump(Formatter *f) const {
       {
@@ -596,6 +597,7 @@ protected:
   struct PullInfo {
     ObjectRecoveryProgress recovery_progress;
     ObjectRecoveryInfo recovery_info;
+    int priority;
 
     void dump(Formatter *f) const {
       {
@@ -616,7 +618,7 @@ protected:
   };
   map<hobject_t, PullInfo> pulling;
 
-  ObjectRecoveryInfo recalc_subsets(ObjectRecoveryInfo recovery_info);
+  ObjectRecoveryInfo recalc_subsets(const ObjectRecoveryInfo& recovery_info);
   static void trim_pushed_data(const interval_set<uint64_t> &copy_subset,
 			       const interval_set<uint64_t> &intervals_received,
 			       bufferlist data_received,
@@ -624,12 +626,12 @@ protected:
 			       bufferlist *data_usable);
   void handle_pull_response(OpRequestRef op);
   void handle_push(OpRequestRef op);
-  int send_push(int peer,
-		ObjectRecoveryInfo recovery_info,
+  int send_push(int priority, int peer,
+		const ObjectRecoveryInfo& recovery_info,
 		ObjectRecoveryProgress progress,
 		ObjectRecoveryProgress *out_progress = 0);
-  int send_pull(int peer,
-		ObjectRecoveryInfo recovery_info,
+  int send_pull(int priority, int peer,
+		const ObjectRecoveryInfo& recovery_info,
 		ObjectRecoveryProgress progress);
   void submit_push_data(const ObjectRecoveryInfo &recovery_info,
 			bool first,
@@ -741,7 +743,8 @@ protected:
   // Reverse mapping from osd peer to objects beging pulled from that peer
   map<int, set<hobject_t> > pull_from_peer;
 
-  int recover_object_replicas(const hobject_t& soid, eversion_t v);
+  int recover_object_replicas(const hobject_t& soid, eversion_t v,
+			      int priority);
   void calc_head_subsets(ObjectContext *obc, SnapSet& snapset, const hobject_t& head,
 			 pg_missing_t& missing,
 			 const hobject_t &last_backfill,
@@ -751,10 +754,16 @@ protected:
 			  const hobject_t &last_backfill,
 			  interval_set<uint64_t>& data_subset,
 			  map<hobject_t, interval_set<uint64_t> >& clone_subsets);
-  void push_to_replica(ObjectContext *obc, const hobject_t& oid, int dest);
-  void push_start(ObjectContext *obc,
+  void push_to_replica(
+    ObjectContext *obc,
+    const hobject_t& oid,
+    int dest,
+    int priority);
+  void push_start(int priority,
+		  ObjectContext *obc,
 		  const hobject_t& oid, int dest);
-  void push_start(ObjectContext *obc,
+  void push_start(int priority,
+		  ObjectContext *obc,
 		  const hobject_t& soid, int peer,
 		  eversion_t version,
 		  interval_set<uint64_t> &data_subset,
@@ -765,7 +774,9 @@ protected:
 
   // Cancels/resets pulls from peer
   void check_recovery_sources(const OSDMapRef map);
-  int pull(const hobject_t& oid, eversion_t v);
+  int pull(
+    const hobject_t& oid, eversion_t v,
+    int priority);
 
   // low level ops
 
@@ -783,7 +794,6 @@ protected:
   int prepare_transaction(OpContext *ctx);
   
   // pg on-disk content
-  void remove_object_with_snap_hardlinks(ObjectStore::Transaction& t, const hobject_t& soid);
   void clean_up_local(ObjectStore::Transaction& t);
 
   void _clear_recovery_state();
@@ -828,7 +838,7 @@ protected:
     list<ObjectStore::Transaction*> tls;
     
     RepModify() : pg(NULL), ctx(NULL), applied(false), committed(false), ackerosd(-1),
-		  bytes_written(0) {}
+		  epoch_started(0), bytes_written(0) {}
   };
 
   struct C_OSD_RepModifyApply : public Context {
@@ -885,6 +895,15 @@ protected:
       pg->put();
     }
   };
+  struct C_OSD_AppliedRecoveredObjectReplica : public Context {
+    boost::intrusive_ptr<ReplicatedPG> pg;
+    ObjectStore::Transaction *t;
+    C_OSD_AppliedRecoveredObjectReplica(ReplicatedPG *p, ObjectStore::Transaction *tt) :
+      pg(p), t(tt) {}
+    void finish(int r) {
+      pg->_applied_recovered_object_replica(t);
+    }
+  };
 
   void sub_op_remove(OpRequestRef op);
 
@@ -894,6 +913,7 @@ protected:
 
   void sub_op_modify_reply(OpRequestRef op);
   void _applied_recovered_object(ObjectStore::Transaction *t, ObjectContext *obc);
+  void _applied_recovered_object_replica(ObjectStore::Transaction *t);
   void _committed_pushed_object(OpRequestRef op, epoch_t same_since, eversion_t lc);
   void recover_got(hobject_t oid, eversion_t v);
   void sub_op_push(OpRequestRef op);
@@ -905,7 +925,10 @@ protected:
 
 
   // -- scrub --
-  virtual int _scrub(ScrubMap& map, int& errors, int& fixed);
+  virtual void _scrub(ScrubMap& map);
+  virtual void _scrub_clear_state();
+  virtual void _scrub_finish();
+  object_stat_collection_t scrub_cstat;
 
   void apply_and_flush_repops(bool requeue);
 
@@ -918,7 +941,7 @@ protected:
 
 public:
   ReplicatedPG(OSDService *o, OSDMapRef curmap,
-	       PGPool _pool, pg_t p, const hobject_t& oid,
+	       const PGPool &_pool, pg_t p, const hobject_t& oid,
 	       const hobject_t& ioid);
   ~ReplicatedPG() {}
 
@@ -937,6 +960,10 @@ public:
   RepGather *trim_object(const hobject_t &coid, const snapid_t &sn);
   void snap_trimmer();
   int do_osd_ops(OpContext *ctx, vector<OSDOp>& ops);
+
+  int do_tmapup(OpContext *ctx, bufferlist::iterator& bp, OSDOp& osd_op);
+  int do_tmapup_slow(OpContext *ctx, bufferlist::iterator& bp, OSDOp& osd_op, bufferlist& bl);
+
   void do_osd_op_effects(OpContext *ctx);
 private:
   bool temp_created;
@@ -1026,6 +1053,8 @@ public:
   bool is_missing_object(const hobject_t& oid);
   void wait_for_missing_object(const hobject_t& oid, OpRequestRef op);
   void wait_for_all_missing(OpRequestRef op);
+  void wait_for_backfill_pos(OpRequestRef op);
+  void release_waiting_for_backfill_pos();
 
   bool is_degraded_object(const hobject_t& oid);
   void wait_for_degraded_object(const hobject_t& oid, OpRequestRef op);
