@@ -392,6 +392,8 @@ std::string pg_state_string(int state)
     oss << "active+";
   if (state & PG_STATE_CLEAN)
     oss << "clean+";
+  if (state & PG_STATE_RECOVERY_WAIT)
+    oss << "recovery_wait+";
   if (state & PG_STATE_RECOVERING)
     oss << "recovering+";
   if (state & PG_STATE_DOWN)
@@ -443,8 +445,16 @@ void pool_snap_info_t::dump(Formatter *f) const
   f->dump_string("name", name);
 }
 
-void pool_snap_info_t::encode(bufferlist& bl) const
+void pool_snap_info_t::encode(bufferlist& bl, uint64_t features) const
 {
+  if ((features & CEPH_FEATURE_PGPOOL3) == 0) {
+    __u8 struct_v = 1;
+    ::encode(struct_v, bl);
+    ::encode(snapid, bl);
+    ::encode(stamp, bl);
+    ::encode(name, bl);
+    return;
+  }
   ENCODE_START(2, 2, bl);
   ::encode(snapid, bl);
   ::encode(stamp, bl);
@@ -659,7 +669,7 @@ void pg_pool_t::encode(bufferlist& bl, uint64_t features) const
 
     ::encode(auid, bl);
 
-    ::encode_nohead(snaps, bl);
+    ::encode_nohead(snaps, bl, features);
     removed_snaps.encode_nohead(bl);
     return;
   }
@@ -679,7 +689,7 @@ void pg_pool_t::encode(bufferlist& bl, uint64_t features) const
     ::encode(last_change, bl);
     ::encode(snap_seq, bl);
     ::encode(snap_epoch, bl);
-    ::encode(snaps, bl);
+    ::encode(snaps, bl, features);
     ::encode(removed_snaps, bl);
     ::encode(auid, bl);
     ::encode(flags, bl);
@@ -687,7 +697,7 @@ void pg_pool_t::encode(bufferlist& bl, uint64_t features) const
     return;
   }
 
-  ENCODE_START(6, 5, bl);
+  ENCODE_START(7, 5, bl);
   ::encode(type, bl);
   ::encode(size, bl);
   ::encode(crush_ruleset, bl);
@@ -700,17 +710,18 @@ void pg_pool_t::encode(bufferlist& bl, uint64_t features) const
   ::encode(last_change, bl);
   ::encode(snap_seq, bl);
   ::encode(snap_epoch, bl);
-  ::encode(snaps, bl);
+  ::encode(snaps, bl, features);
   ::encode(removed_snaps, bl);
   ::encode(auid, bl);
   ::encode(flags, bl);
   ::encode(crash_replay_interval, bl);
+  ::encode(min_size, bl);
   ENCODE_FINISH(bl);
 }
 
 void pg_pool_t::decode(bufferlist::iterator& bl)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(6, 5, 5, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(7, 5, 5, bl);
   ::decode(type, bl);
   ::decode(size, bl);
   ::decode(crush_ruleset, bl);
@@ -754,6 +765,11 @@ void pg_pool_t::decode(bufferlist::iterator& bl)
     else
       crash_replay_interval = 0;
   }
+  if (struct_v >= 7) {
+    ::decode(min_size, bl);
+  } else {
+    min_size = size - size/2;
+  }
   DECODE_FINISH(bl);
   calc_pg_masks();
 }
@@ -1270,7 +1286,7 @@ void pg_history_t::generate_test_instances(list<pg_history_t*>& o)
 
 void pg_info_t::encode(bufferlist &bl) const
 {
-  ENCODE_START(26, 26, bl);
+  ENCODE_START(27, 26, bl);
   ::encode(pgid, bl);
   ::encode(last_update, bl);
   ::encode(last_complete, bl);
@@ -1279,12 +1295,13 @@ void pg_info_t::encode(bufferlist &bl) const
   ::encode(stats, bl);
   history.encode(bl);
   ::encode(purged_snaps, bl);
+  ::encode(last_epoch_started, bl);
   ENCODE_FINISH(bl);
 }
 
 void pg_info_t::decode(bufferlist::iterator &bl)
 {
-  DECODE_START_LEGACY_COMPAT_LEN(26, 26, 26, bl);
+  DECODE_START_LEGACY_COMPAT_LEN(27, 26, 26, bl);
   if (struct_v < 23) {
     old_pg_t opgid;
     ::decode(opgid, bl);
@@ -1309,6 +1326,11 @@ void pg_info_t::decode(bufferlist::iterator &bl)
     set<snapid_t> snap_trimq;
     ::decode(snap_trimq, bl);
   }
+  if (struct_v < 27) {
+    last_epoch_started = history.last_epoch_started;
+  } else {
+    ::decode(last_epoch_started, bl);
+  }
   DECODE_FINISH(bl);
 }
 
@@ -1332,6 +1354,7 @@ void pg_info_t::dump(Formatter *f) const
   f->dump_int("empty", is_empty());
   f->dump_int("dne", dne());
   f->dump_int("incomplete", is_incomplete());
+  f->dump_int("last_epoch_started", last_epoch_started);
 }
 
 void pg_info_t::generate_test_instances(list<pg_info_t*>& o)
@@ -1454,18 +1477,23 @@ bool pg_interval_t::check_new_interval(
   epoch_t last_epoch_clean,
   OSDMapRef osdmap,
   OSDMapRef lastmap,
+  int64_t pool_id,
   map<epoch_t, pg_interval_t> *past_intervals,
   std::ostream *out)
 {
   // remember past interval
-  if (new_acting != old_acting || new_up != old_up) {
+  if (new_acting != old_acting || new_up != old_up ||
+      (!(lastmap->get_pools().count(pool_id))) ||
+      lastmap->get_pools().find(pool_id)->second.min_size !=
+      osdmap->get_pools().find(pool_id)->second.min_size) {
     pg_interval_t& i = (*past_intervals)[same_interval_since];
     i.first = same_interval_since;
     i.last = osdmap->get_epoch() - 1;
     i.acting = old_acting;
     i.up = old_up;
 
-    if (i.acting.size()) {
+    if (i.acting.size() >=
+	osdmap->get_pools().find(pool_id)->second.min_size) {
       if (lastmap->get_up_thru(i.acting[0]) >= i.first &&
 	  lastmap->get_up_from(i.acting[0]) <= i.first) {
 	i.maybe_went_rw = true;
@@ -1579,15 +1607,30 @@ void pg_query_t::generate_test_instances(list<pg_query_t*>& o)
 
 void pg_log_entry_t::encode(bufferlist &bl) const
 {
-  ENCODE_START(5, 4, bl);
+  ENCODE_START(6, 4, bl);
   ::encode(op, bl);
   ::encode(soid, bl);
   ::encode(version, bl);
-  ::encode(prior_version, bl);
+
+  /**
+   * Added with reverting_to:
+   * Previous code used prior_version to encode
+   * what we now call reverting_to.  This will
+   * allow older code to decode reverting_to
+   * into prior_version as expected.
+   */
+  if (op == LOST_REVERT)
+    ::encode(reverting_to, bl);
+  else
+    ::encode(prior_version, bl);
+
   ::encode(reqid, bl);
   ::encode(mtime, bl);
   if (op == CLONE)
     ::encode(snaps, bl);
+
+  if (op == LOST_REVERT)
+    ::encode(prior_version, bl);
   ENCODE_FINISH(bl);
 }
 
@@ -1607,13 +1650,27 @@ void pg_log_entry_t::decode(bufferlist::iterator &bl)
   if (struct_v < 3)
     invalid_hash = true;
   ::decode(version, bl);
-  ::decode(prior_version, bl);
+
+  if (struct_v >= 6 && op == LOST_REVERT)
+    ::decode(reverting_to, bl);
+  else
+    ::decode(prior_version, bl);
+
   ::decode(reqid, bl);
   ::decode(mtime, bl);
   if (op == CLONE)
     ::decode(snaps, bl);
   if (struct_v < 5)
     invalid_pool = true;
+
+  if (op == LOST_REVERT) {
+    if (struct_v >= 6) {
+      ::decode(prior_version, bl);
+    } else {
+      reverting_to = prior_version;
+    }
+  }
+
   DECODE_FINISH(bl);
 }
 
@@ -2677,6 +2734,7 @@ ostream& operator<<(ostream& out, const OSDOp& op)
   if (ceph_osd_op_type_data(op.op.op)) {
     // data extent
     switch (op.op.op) {
+    case CEPH_OSD_OP_STAT:
     case CEPH_OSD_OP_DELETE:
       break;
     case CEPH_OSD_OP_TRUNCATE:
