@@ -57,6 +57,12 @@ struct ObjectOperation {
   vector<int*> out_rval;
 
   ObjectOperation() : flags(0), priority(0) {}
+  ~ObjectOperation() {
+    while (!out_handler.empty()) {
+      delete out_handler.back();
+      out_handler.pop_back();
+    }
+  }
 
   size_t size() {
     return ops.size();
@@ -615,6 +621,7 @@ public:
     utime_t stamp;
 
     bool precalc_pgid;
+    epoch_t map_dne_bound;
 
     bool budgeted;
 
@@ -631,6 +638,7 @@ public:
       flags(f), priority(0), onack(ac), oncommit(co),
       tid(0), attempts(0),
       paused(false), objver(ov), reply_epoch(NULL), precalc_pgid(false),
+      map_dne_bound(0),
       budgeted(false),
       should_resend(true) {
       ops.swap(op);
@@ -648,6 +656,12 @@ public:
       if (oloc.key == o)
 	oloc.key.clear();
     }
+    ~Op() {
+      while (!out_handler.empty()) {
+	delete out_handler.back();
+	out_handler.pop_back();
+      }
+    }
 
     bool operator<(const Op& other) const {
       return tid < other.tid;
@@ -657,7 +671,8 @@ public:
   struct C_Op_Map_Latest : public Context {
     Objecter *objecter;
     tid_t tid;
-    C_Op_Map_Latest(Objecter *o, tid_t t) : objecter(o), tid(t) {}
+    version_t latest;
+    C_Op_Map_Latest(Objecter *o, tid_t t) : objecter(o), tid(t), latest(0) {}
     void finish(int r);
   };
 
@@ -800,12 +815,14 @@ public:
     xlist<LingerOp*>::item session_item;
 
     tid_t register_tid;
+    epoch_t map_dne_bound;
 
     LingerOp() : linger_id(0), flags(0), poutbl(NULL), pobjver(NULL),
 		 registered(false),
 		 on_reg_ack(NULL), on_reg_commit(NULL),
 		 session(NULL), session_item(this),
-		 register_tid(0) {}
+		 register_tid(0),
+		 map_dne_bound(0) {}
 
     // no copy!
     const LingerOp &operator=(const LingerOp& r);
@@ -845,8 +862,9 @@ public:
   struct C_Linger_Map_Latest : public Context {
     Objecter *objecter;
     uint64_t linger_id;
+    version_t latest;
     C_Linger_Map_Latest(Objecter *o, uint64_t id) :
-      objecter(o), linger_id(id) {}
+      objecter(o), linger_id(id), latest(0) {}
     void finish(int r);
   };
 
@@ -895,9 +913,11 @@ public:
   void _linger_ack(LingerOp *info, int r);
   void _linger_commit(LingerOp *info, int r);
 
-  void op_check_for_latest_map(Op *op);
+  void check_op_pool_dne(Op *op);
+  void _send_op_map_check(Op *op);
   void op_cancel_map_check(Op *op);
-  void linger_check_for_latest_map(LingerOp *op);
+  void check_linger_pool_dne(LingerOp *op);
+  void _send_linger_map_check(LingerOp *op);
   void linger_cancel_map_check(LingerOp *op);
 
   void kick_requests(OSDSession *session);
@@ -961,8 +981,10 @@ public:
     assert(!logger);
   }
 
-  void init();
-  void shutdown();
+  void init_unlocked();
+  void init_locked();
+  void shutdown_locked();
+  void shutdown_unlocked();
 
   /**
    * Tell the objecter to throttle outgoing ops according to its
@@ -1490,7 +1512,7 @@ public:
       C_GatherBuilder gcom(cct, oncommit);
       for (vector<ObjectExtent>::iterator p = extents.begin(); p != extents.end(); p++) {
 	bufferlist cur;
-	for (map<uint64_t, uint64_t>::iterator bit = p->buffer_extents.begin();
+	for (vector<pair<uint64_t,uint64_t> >::iterator bit = p->buffer_extents.begin();
 	     bit != p->buffer_extents.end();
 	     bit++)
 	  bl.copy(bit->first, bit->second, cur);
