@@ -9,7 +9,10 @@
 #ifndef CEPH_RGW_OP_H
 #define CEPH_RGW_OP_H
 
+#include <limits.h>
+
 #include <string>
+#include <map>
 
 #include "rgw_common.h"
 #include "rgw_rados.h"
@@ -22,6 +25,8 @@ struct req_state;
 class RGWHandler;
 
 void rgw_get_request_metadata(struct req_state *s, map<string, bufferlist>& attrs);
+int rgw_build_policies(RGWRados *store, struct req_state *s, bool only_bucket, bool prefetch_data);
+
 
 /**
  * Provide the base class for all ops.
@@ -108,7 +113,7 @@ public:
   int handle_user_manifest(const char *prefix);
 
   virtual int get_params() = 0;
-  virtual int send_response(bufferlist& bl) = 0;
+  virtual int send_response_data(bufferlist& bl) = 0;
 
   virtual const char *name() { return "get_obj"; }
 };
@@ -249,8 +254,18 @@ class RGWPutObjProcessor
 protected:
   RGWRados *store;
   struct req_state *s;
+  bool is_complete;
+
+  virtual int do_complete(string& etag, map<string, bufferlist>& attrs) = 0;
+
+  list<rgw_obj> objs;
+
+  void add_obj(rgw_obj& obj) {
+    objs.push_back(obj);
+  }
 public:
-  virtual ~RGWPutObjProcessor() {}
+  RGWPutObjProcessor() : store(NULL), s(NULL), is_complete(false) {}
+  virtual ~RGWPutObjProcessor();
   virtual int prepare(RGWRados *_store, struct req_state *_s) {
     store = _store;
     s = _s;
@@ -258,7 +273,7 @@ public:
   };
   virtual int handle_data(bufferlist& bl, off_t ofs, void **phandle) = 0;
   virtual int throttle_data(void *handle) = 0;
-  virtual int complete(string& etag, map<string, bufferlist>& attrs) = 0;
+  virtual int complete(string& etag, map<string, bufferlist>& attrs);
 };
 
 class RGWPutObj : public RGWOp {
@@ -302,6 +317,55 @@ public:
   virtual const char *name() { return "put_obj"; }
 };
 
+class RGWPostObj : public RGWOp {
+
+  friend class RGWPutObjProcessor;
+
+protected:
+  off_t min_len;
+  off_t max_len;
+  int ret;
+  int len;
+  off_t ofs;
+  const char *supplied_md5_b64;
+  const char *supplied_etag;
+  string etag;
+  string boundary;
+  bool data_pending;
+  string content_type;
+  RGWAccessControlPolicy policy;
+  map<string, bufferlist> attrs;
+
+public:
+  RGWPostObj() {}
+
+  virtual void init(RGWRados *store, struct req_state *s, RGWHandler *h) {
+    RGWOp::init(store, s, h);
+    min_len = 0;
+    max_len = LLONG_MAX;
+    ret = 0;
+    len = 0;
+    ofs = 0;
+    supplied_md5_b64 = NULL;
+    supplied_etag = NULL;
+    etag = "";
+    boundary = "";
+    data_pending = false;
+    policy.set_ctx(s->cct);
+  }
+
+  int verify_permission();
+  void execute();
+
+  RGWPutObjProcessor *select_processor();
+  void dispose_processor(RGWPutObjProcessor *processor);
+
+  virtual int get_params() = 0;
+  virtual int get_data(bufferlist& bl) = 0;
+  virtual void send_response() = 0;
+  virtual const char *name() { return "post_obj"; }
+};
+
 class RGWPutMetadata : public RGWOp {
 protected:
   int ret;
@@ -429,6 +493,9 @@ public:
     len = 0;
     data = NULL;
   }
+  virtual ~RGWPutACLs() {
+    free(data);
+  }
 
   int verify_permission();
   void execute();
@@ -476,6 +543,9 @@ public:
     data = NULL;
     len = 0;
   }
+  virtual ~RGWCompleteMultipart() {
+    free(data);
+  }
 
   int verify_permission();
   void execute();
